/******************************************************************************
 *
 * Copyright (c) 1999-2003 AppGate Network Security AB. All Rights Reserved.
 * 
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the MindTerm Public Source License,
 * Version 2.0, (the 'License'). You may not use this file except in compliance
 * with the License.
 * 
 * You should have received a copy of the MindTerm Public Source License
 * along with this software; see the file LICENSE.  If not, write to
 * AppGate Network Security AB, Otterhallegatan 2, SE-41118 Goteborg, SWEDEN
 *
 *****************************************************************************/

package com.mindbright.examples;

import java.io.InputStream;
import java.io.OutputStream;

import java.net.Socket;
import java.net.ServerSocket;
import java.net.InetAddress;

import com.mindbright.ssh2.SSH2Transport;
import com.mindbright.ssh2.SSH2SimpleClient;
import com.mindbright.ssh2.SSH2FTPOverSFTP;
import com.mindbright.ssh2.SSH2Preferences;

import com.mindbright.util.SecureRandomAndPad;
import com.mindbright.util.RandomSeed;
import com.mindbright.jca.security.SecureRandom;

/**
 * This is a demo of one usage of the FTP over SFTP bridging capability. It
 * starts a local server which acts like an ftp server except that the ftp
 * username is used to address different ssh2 servers (with sftp enabled) to
 * connect to.  The format of the ftp username is:
 * <p>
 * <pre>
 * ssh2-user@ssh2-host[:port]
 * </pre>
 * <p>
 * <pre>
 * usage: FTPToSFTPProxy [&lt;listen-address&gt;[:&lt;port&gt;]]
 * </pre>
 */
public class FTPToSFTPProxy extends SSH2FTPOverSFTP {

    public final static String IDENTITY =
	"FTPToSFTPProxy use 'ssh2-user@ssh2-host[:port]' as your username.";

    private Socket             ssh2Socket;
    private SSH2Transport      transport;
    private SSH2SimpleClient   client;
    private SecureRandomAndPad secureRandom;


    public FTPToSFTPProxy(InputStream ftpInput, OutputStream ftpOutput) {
	super(ftpInput, ftpOutput, IDENTITY, true);
    }

    public boolean login(String user, String pass) {
	int i = user.indexOf('@');
	if(i == -1) {
	    return false;
	}
	String ssh2Server   = user.substring(i + 1);
	int    port     = 22;
	String ssh2User = user.substring(0, i);
	i = ssh2Server.indexOf(':');
	if(i != -1) {
	    String p = ssh2Server.substring(i + 1);
	    ssh2Server = ssh2Server.substring(0, i);
	    port   = Integer.parseInt(p);
	}
	try {
	    /*
	     * TODO: Should include code to read host key fingerprints from a
	     * file and set a SSH2HostKeyVerifier here to avoid MITM attack.
	     */
	    SSH2Preferences prefs = new SSH2Preferences();
	    prefs.setPreference(SSH2Preferences.LOG_LEVEL, "4");
	    ssh2Socket = new Socket(ssh2Server, port);
	    transport  = new SSH2Transport(ssh2Socket, prefs,
					   createSecureRandom());
	    client     = new SSH2SimpleClient(transport, ssh2User, pass);
	    initSFTP(client.getConnection());
	    return super.login(user, pass);
	} catch (Exception e) {
	    return false;
	}
    }

    public void quit() {
	if(connection != null) {
	    super.quit();
	}
	sftp.terminate();
    }

    public SecureRandomAndPad createSecureRandom() {
	if(secureRandom == null) {
	    RandomSeed seed = new RandomSeed("/dev/random", "/dev/urandom");
	    byte[] s = seed.getBytesBlocking(20);
	    secureRandom = new SecureRandomAndPad(new SecureRandom(s));
	}
	return secureRandom;
    }

    public static void main(String[] argv) {
	if(argv.length > 1) {
	    System.out.println("usage: FTPToSFTPProxy [<listen-address>[:<port>]]");
	    System.exit(1);
	}
	String server = "127.0.0.1";
	int    port   = 21;
	try {
	    if(argv.length > 0) {
		server = argv[0];
		int i = server.indexOf(':');
		if(i != -1) {
		    String p = server.substring(i + 1);
		    server = server.substring(0, i);
		    port   = Integer.parseInt(p);
		}
	    }
	    System.out.println("** FTPToSFTPProxy listening on " + server + ":"
			       + port);
	    ServerSocket listen =
		new ServerSocket(port, 32, InetAddress.getByName(server));
	    while(true) {
		Socket conn = listen.accept();
		FTPToSFTPProxy proxy =
		    new FTPToSFTPProxy(conn.getInputStream(),
				       conn.getOutputStream());
	    }
	} catch (Exception e) {
	    System.out.println("An error occured: " + e.getMessage());
	}
    }

}
