/******************************************************************************
 *
 * Copyright (c) 1999-2003 AppGate Network Security AB. All Rights Reserved.
 * 
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the MindTerm Public Source License,
 * Version 2.0, (the 'License'). You may not use this file except in compliance
 * with the License.
 * 
 * You should have received a copy of the MindTerm Public Source License
 * along with this software; see the file LICENSE.  If not, write to
 * AppGate Network Security AB, Otterhallegatan 2, SE-41118 Goteborg, SWEDEN
 *
 *****************************************************************************/

package com.mindbright.examples;

import java.io.*;
import java.net.*;
import java.util.*;
import com.mindbright.ssh.*;
import com.mindbright.sshcommon.*;
import com.mindbright.jca.security.interfaces.RSAPublicKey;

/**
 * Examples on how to use the ssh1 classes.
 *
 * <pre>
 *     - runWithPortFwd: starts SSH1 connection and port forwarding which
 *        is used to send and receive data over
 *
 *     - runWithCommand: starts a command over an SSH1 connection and
 *        retrieves the output
 *
 *     - runWithCommandInShell: start a shell over SSH1 connection and 
 *        run a command in the shell and retrieve the output
 *
 *     - runSCP: copy a file with SCP
 *
 * Run with: java SSH1Example <host> <port> <username> <pass>
 * </pre>
 */
public class SSH1Example extends SSHInteractorAdapter
    implements SSHAuthenticator, SSHClientUser 
{
    // Modify this to point to a webserver
    private final static String A_WebHost = "localhost";
    
    private Properties props = new Properties();

    private MYSSHClient client;

    class MYSSHClient extends SSHClient 
    {
        MYSSHClient(SSHAuthenticator authenticator, SSHClientUser user) {
            super(authenticator, user);
        }

        boolean myIsCipherSupported(int c) {
            int cipherMask = (0x01 << c);
            if((cipherMask & supportedCiphers) != 0)
                return true;
            return false;
        }
    }

    SSH1Example(String host, String port, String user, String pass) {
        // Save properties for later use in methods from
        // the SSHAuthenticator and SSHClientUser interfaces
        props.put("server", host);
        props.put("port", port);
        props.put("username", user);
        props.put("password", pass);
        props.put("auth-method", "password");
    }

    private final static void listThreads() {
	Thread t = Thread.currentThread();
	ThreadGroup tmp, tg = t.getThreadGroup();
	while ((tmp=tg.getParent()) != null)
	    tg = tmp;
	Thread ts[] = new Thread[tg.activeCount()];
	tg.enumerate(ts, true);
	for (int i=0; i<ts.length; i++) {
	    System.out.println(i+"="+ts[i]);
	}
    }

    /**
     * Start an SSH1 connection with port forward. Then demonstrate
     * reading and writing on this portforward.
     */
    public void runWithPortFwd() throws UnknownHostException, IOException {
	// Prepare the client
        client = new MYSSHClient(this, this);

	/*
	 * Add the port forward from localhost:4083 to "a_WebHost"
	 * (contant defined above) port 80. The portforward is of type
	 * "general" (=no plugin)
	 */
        client.addLocalPortForward(4083, A_WebHost, 80, "general");

	/*
	 * Boot the ssh protocl (cases it to connect).
	 */
        client.bootSSH(false);

	/*
	 * Open a TCP connection to the local end of the portforward.
	 */
        Socket s = new Socket("localhost", 4083);
        OutputStream os = s.getOutputStream();

	// Send a simple web request
        os.write("GET / HTTP/1.0\r\n\r\n".getBytes());
        os.flush();

	// Get reply
        System.out.println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
        BufferedReader in =
	    new BufferedReader(new InputStreamReader(s.getInputStream()));
        for (;;) {
            String line = in.readLine();
            if (line == null)
                break;
            System.out.println(line);
        }
        System.out.println("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
        s.close();

	// Close everything
        client.forcedDisconnect();
        client.clearAllForwards();
    }


    /**
     * Start a command over SSH1 connection and run a command and
     * retrieve the output
     *
     * @param command Command to execute
     */
    public void runWithCommand(String command) throws IOException {
	// Create a suitable client instance
        SSHConsoleClient client =
	    new SSHConsoleClient(getSrvHost(), getSrvPort(), this, this);

	// Register the command
        client.command(command);

	// Show output
        System.out.println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
        BufferedReader in =
	    new BufferedReader(new InputStreamReader(client.getStdOut()));
        for (;;) {
            String line = in.readLine();
            if (line == null)
                break;
            System.out.println(line);
        }
        System.out.println("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
    }


    /**
     * Start a shell over SSH1 connection and run a command and
     * retrieve the output
     *
     * @param command Command to execute
     */
    // Example: run a command over SSH1 connection and retrieve the output
    public void runWithCommandInShell(String command) throws IOException {
	// Create a suitable client instance
        SSHConsoleClient client =
	    new SSHConsoleClient(getSrvHost(), getSrvPort(), this, this);

	// Ask the client to start a shell on the server
        client.shell();

	// Send command to stdin of shell
        OutputStream os = client.getStdIn();
        os.write((command + "; exit\n").getBytes());
        os.flush();

	// Show output
        System.out.println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
        BufferedReader in =
	    new BufferedReader(new InputStreamReader(client.getStdOut()));
        for (;;) {
            String line = in.readLine();
            if (line == null)
                break;
            System.out.println(line);
        }
        System.out.println("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
    }


    /**
     * Transfer files with SCP
     *
     * @param remotefile File on remote system to transfer
     */
    public void runSCP(String remotefile) throws IOException {
        SSHConsoleClient client;
        SSHSCP1 scp1;
        
        System.out.println("Copying file from remote host");
        client = new SSHConsoleClient(getSrvHost(), getSrvPort(), this, this);
        scp1 = new SSHSCP1(new File("."), client, true);
        scp1.copyToLocal("dummy", remotefile, false);
        client.close();

        System.out.println("Copying file to remote host");
        client = new SSHConsoleClient(getSrvHost(), getSrvPort(), this, this);
        scp1 = new SSHSCP1(new File("."), client, true);
        scp1.copyToRemote("dummy", "dummy2", false);        
        client.close();

        (new File("dummy")).delete();
    }


    private String getProp(String key) {
        return (String)props.get(key);
    }


    ////////////////////////////////////////////////////////////////////////
    // SSHAuthenticator

    public String getUsername(SSHClientUser origin) throws IOException {
        return getProp("username");
    }

    public String getPassword(SSHClientUser origin) throws IOException {
        return getProp("password");
    }
    
    public String getChallengeResponse(SSHClientUser origin, String challenge)
	throws IOException {
        return null;
    }
    
    public int[] getAuthTypes(SSHClientUser origin) {
        return SSH.getAuthTypes(getProp("auth-method"));
    }
    
    private static int[] ciphers = {
        SSH.CIPHER_BLOWFISH,
        SSH.CIPHER_IDEA,
        SSH.CIPHER_3DES,
        SSH.CIPHER_RC4,
        SSH.CIPHER_DES
    };
    
    public int getCipher(SSHClientUser origin) {
        for (int i=0; i<ciphers.length; i++)
            if (client.myIsCipherSupported(ciphers[i]))
                return ciphers[i];
        return SSH.CIPHER_DEFAULT;
    }
    
    public SSHRSAKeyFile getIdentityFile(SSHClientUser origin)
	throws IOException {
        return null;
    }
    
    public String getIdentityPassword(SSHClientUser origin)
	throws IOException {
        return null;
    }
    
    public boolean verifyKnownHosts(RSAPublicKey hostPub) throws IOException {
        // TODO:
        return true;
    }

    //
    ////////////////////////////////////////////////////////////////////////


    
    ////////////////////////////////////////////////////////////////////////
    // SSHClientUser
    
    public String getSrvHost() throws IOException { return getProp("server"); }
    public int getSrvPort() { return Integer.parseInt(getProp("port")); }
    public Socket getProxyConnection() throws IOException { return null; }
    public String getDisplay()           { return null; }
    public int getMaxPacketSz()          { return 0; }
    public int getAliveInterval()        { return 0; }
    public int getCompressionLevel()     { return 0; }
    public boolean wantX11Forward()      { return false; }
    public boolean wantPTY()             { return false; }
    public SSHInteractor getInteractor() { return this; }
    
    //
    ////////////////////////////////////////////////////////////////////////


    public static void main(String argv[]) {
        if (argv == null || argv.length != 4) {
            System.out.println("Usage: java SSH1Example <host> <port> <user> <pass>");
            return;
        }
        
        try {
            SSH1Example ssh1 =
		new SSH1Example(argv[0], argv[1], argv[2], argv[3]);
            ssh1.runWithPortFwd();
	    ssh1.runWithCommand("/bin/ls");
	    ssh1.runWithCommandInShell("/bin/ls");
	    ssh1.runSCP(".cshrc"); // Must exist - change if needed.

        } catch (Throwable t) {
            t.printStackTrace();
        }
    }
}
