/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import uk.ac.standrews.cs.nds.util.ClassPath;
import uk.ac.standrews.cs.nds.util.PlatformDescriptor;
import uk.ac.standrews.cs.nds.util.StringConstants;

/**
 * Reifies some aspects of the project, in particular its required classpath.
 *
 * @author graham
 */
public class ProjectMetaData {

	private static final String PACKAGE_SEPARATOR = ".";
	private static final String PACKAGE_SEPARATOR_PATTERN = "\\.";
	private static final String PACKAGE_PREFIX = "uk" + PACKAGE_SEPARATOR + "ac" + PACKAGE_SEPARATOR + "standrews" + PACKAGE_SEPARATOR + "cs";
	private static ProjectMetaData singleton_instance = new ProjectMetaData();

	// ***********************************************************************

	/**
	 * @return the metadata for this project
	 */
	public static ProjectMetaData getInstance() {

		return singleton_instance;
	}

	/**
	 * Returns the metadata for the project containing the given class.
	 *
	 * @param clazz a class
	 * @return the metadata for the project containing the given class
	 * @throws ClassNotFoundException if there is no project metadata for the given class
	 */
	public static ProjectMetaData getProject(Class<?> clazz) throws ClassNotFoundException {

		throw new ClassNotFoundException();
	}

	// ***********************************************************************

	/**
	 * @return the name of this project
	 */
	private List<String> getProjectNames() {

		List<String> project_names = new ArrayList<String>();

		Class<?> c = getClass();

		try {
			while (c != null) {

				project_names.add(getProjectName(c));
				c = c.getSuperclass();
			}
		}
		catch (ClassNotFoundException e) { /* Finished. */ }

		return project_names;
	}

	protected String getProjectName(Class<?> clazz) throws ClassNotFoundException {

		Package this_package = clazz.getPackage();

		String full_package_name = this_package.getName();
		String s = PACKAGE_PREFIX;
		if (!full_package_name.startsWith(s)) throw new ClassNotFoundException();

		String[] package_names = full_package_name.split(PACKAGE_SEPARATOR_PATTERN);

		return package_names[4];   // Assume project name is prefixed with 'uk.ac.standrews.cs'.
	}

	/**
	 * Returns the class path needed to execute classes in this project.
	 * @return the current class path of the executing program
	 */
	public ClassPath classPath() {

		// Return the current runtime class path.
		return ClassPath.getCurrentClassPath();
	}

	/**
	 * Returns the class path needed to execute classes in this project on the given platform.
	 * This assumes that the necessary classes are remotely installed at the specified path.
	 *
	 * @param platform a platform descriptor
	 * @param code_base_path the path where the necessary classes are installed
	 * @return the class path for this project on the given platform
	 */
	public ClassPath classPath(PlatformDescriptor platform, File code_base_path) {

		ClassPath result = getProjectJars();
		result.append(getLibraryJars(platform));

		result.resolveRelativePaths(code_base_path);

		return result;
	}

	/**
	 * Returns the class path needed to execute classes in this project on the given platform.
	 * This assumes that the necessary classes are remotely installed at the specified path.
	 *
	 * @param platform a platform descriptor
	 * @param code_base_path the path where the necessary classes are installed
	 * @return the class path for this project on the given platform
	 */
	public ClassPath classPath(PlatformDescriptor platform, File code_base_path, String java_version) {

		ClassPath result = getProjectJars();
		result.append(getLibraryJars(platform));

		result.resolveRelativePaths(code_base_path);

		return result;
	}

	// ***********************************************************************

	// Relative paths, fixed relative to code base, independent of platform.
	private ClassPath getProjectJars() {

		ClassPath class_path = new ClassPath();

		for (String project_name : getProjectNames())
			class_path.append(new File(new File(new File(project_name), "bin"), project_name + ".jar"));

		return class_path;
	}

	// Some platform dependent, some relative & some absolute.
	protected ClassPath getLibraryJars(PlatformDescriptor platform) {

		ClassPath class_path = new ClassPath();
		class_path.append(new File(StringConstants.getString("JUNIT_JAR_PATH")));
		return class_path;
	}
}
