/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.p2p.impl;

import java.math.BigInteger;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;

/**
 * Implementation of key.
 *
 * @author stuart, al, graham, sja7
 */
public class Key implements IKey {

	private static final BigInteger ONE_HUNDRED = new BigInteger("100");

	// ***********************************************************************

	/**
	 * The length of a key in bits.
	 */
	public static final int KEY_LENGTH = 160;

	/**
	 * Big integer representation of two.
	 */
	public static final BigInteger TWO = BigInteger.ONE.add(BigInteger.ONE);

	/**
	 * The size of the keyspace.
	 */
	public static final BigInteger KEYSPACE_SIZE = TWO.pow(KEY_LENGTH);

	/**
	 * The default radix used in converting a key's value to a string.
	 */
	public static final int DEFAULT_RADIX = 16;

	// ***********************************************************************

	private BigInteger key_value;

	// ***********************************************************************

	/**
	 * This is needed for RRT deserialization when sent by value, since the public constructors all fail when passed null.
	 */
	protected Key() {

		// Do nothing.
	}

	// ***********************************************************************

	/**
	 * Creates a new key using the given value modulo the key space size.
	 *
	 * @param key_value the value of the key
	 */
	public Key(BigInteger key_value) {

		this.key_value = key_value.remainder(KEYSPACE_SIZE);

		// Allow for negative key value.
		if (this.key_value.compareTo(BigInteger.ZERO) < 0) {
			this.key_value = this.key_value.add(KEYSPACE_SIZE);
		}
	}

	/**
	 * Creates a new key using a string representation of a BigInteger with radix {@link #DEFAULT_RADIX}.
	 *
	 * @param string_representation a representation of the key value
	 */
	public Key(String string_representation) {

		this(new BigInteger(string_representation, DEFAULT_RADIX));
	}

	/**
	 * Creates a new key using a byte array representation of a BigInteger with radix {@link #DEFAULT_RADIX}.
	 *
	 * @param bytes a representation of the key value
	 */
	public Key(byte[] bytes) {

		this(Radix.toHexString(bytes));
	}

	// ***********************************************************************

	public BigInteger keyValue() {

		return key_value;
	}

	public String toString(int radix) {

		int bits_per_digit = Radix.bitsPerDigit(radix);
		int pad_to_length = KEY_LENGTH / bits_per_digit;

		return toString(radix, pad_to_length);
	}

	public int countSharedPrefixDigits(IKey key, int radix) {

		String representation1 = toString(radix);
		String representation2 = key.toString(radix);

		int shared_digit_count = 0;
		int key_length = representation1.length();

		while (shared_digit_count < key_length && representation1.charAt(shared_digit_count) == representation2.charAt(shared_digit_count)) {
			shared_digit_count++;
		}

		return shared_digit_count;
	}

	public char digit(int index, int radix) throws IndexOutOfBoundsException {

		return toString(radix).charAt(index);
	}

	public String toStringAsKeyspaceFraction() {

		return key_value.multiply(ONE_HUNDRED).divide(KEYSPACE_SIZE) + "%";
	}

	public int bitLength() {

		return key_value.bitLength();
	}

	public int maxBitLength() {

		return KEYSPACE_SIZE.bitLength();
	}

	public BigInteger ringDistanceTo(IKey end) {

		BigInteger distance = end.keyValue().subtract(key_value);

		// Check for ring wrap-around, indicated by negative distance.
		if (distance.compareTo(BigInteger.ZERO) < 0)

			// The segment wraps around, so add the keyspace size.
			return distance.add(KEYSPACE_SIZE);

		else
			// No wrap-around.
			return distance;
	}

	public boolean firstNumericallyCloserThanSecond(IKey first, IKey second) {

		BigInteger distance_to_first =  key_value.subtract(first.keyValue()).abs();
		BigInteger distance_to_second = key_value.subtract(second.keyValue()).abs();

		return distance_to_first.compareTo(distance_to_second) < 0;
	}

	public boolean firstCloserInRingThanSecond(IKey first, IKey second) {

		BigInteger distance_to_first =  ringDistanceTo(first);
		BigInteger distance_to_second = ringDistanceTo(second);

		return distance_to_first.compareTo(distance_to_second) < 0;
	}

	// ***********************************************************************

	public int compareTo(IKey k) {

		return key_value.compareTo(k.keyValue());
	}

	// ***********************************************************************

	/**
	 * Returns a string representation of the key value with radix {@link #DEFAULT_RADIX}.
	 *
	 * @return a string representation of the key value
	 */
	@Override
	public String toString() {

		return toString(DEFAULT_RADIX);
	}

	@Override
	public boolean equals(Object o) {

		try {
			IKey k = (IKey) o;

			return key_value.equals(k.keyValue());

		}
		catch (ClassCastException e) { return false; }
	}

	@Override
	public int hashCode() {

		return toString().hashCode();
	}

	// ***********************************************************************

	private String toString(int radix, int pad_to_length) {

		StringBuffer result = new StringBuffer(key_value.toString(radix));
		while (result.length() < pad_to_length) {
			result.insert(0, '0');
		}
		return result.toString();
	}
}