/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Created on Jan 17, 2005 at 3:56:58 PM.
 */
package uk.ac.standrews.cs.nds.p2p.impl;

/**
 * @author al
 */
public class Radix {

	/**
	 * Returns the value of a given digit in a given radix.
	 *
	 * @param digit the digit
	 * @param radix the radix
	 * @return the corresponding value
	 */
	public static int digitToValue(char digit, int radix) {

		if (digit < 'a') {

			if (digit < '0' || digit > '9') throw new IllegalArgumentException("illegal digit " + digit + " for radix " + radix);
			return digit - '0';
		}
		else {

			if (digit > 'z') throw new IllegalArgumentException("illegal digit " + digit + " for radix " + radix);
			return digit - 'a' + '9' - '0' + 1;
		}
	}

	/**
	 * Returns the digit corresponding to a given value in a given radix.
	 *
	 * @param value the value
	 * @param radix the radix
	 * @return the corresponding digit
	 */
	public static char valueToDigit(int value, int radix) {

		if (value < 0 || value >= radix) throw new IllegalArgumentException("illegal digit " + value + " for radix " + radix);

		if (value < '9' - '0' + 1)
			return (char)('0' + value);
		else
			return (char)('a' + value - ('9' - '0' + 1));
	}

	/**
	 * Returns the number of bits needed to represent a given number in binary.
	 *
	 * @param n the number
	 * @return the number of bits needed to represent the number in binary
	 */
	public static int bitsInBinaryRepresentation(int n) {

		double double_value = Integer.valueOf(n).doubleValue();

		return (int)Math.floor(logBase2(double_value)) + 1;
	}

	/**
	 * Returns the number of bits needed to represent a digit using the given radix.
	 *
	 * @param radix the number
	 * @return the number of bits needed to represent a digit
	 */
	public static int bitsPerDigit(int radix) {

		return bitsInBinaryRepresentation(radix - 1);
	}

	/**
	 * Returns the logarithm of a given number to base 2.
	 *
	 * @param n the number
	 * @return the log of the number to base 2
	 */
	public static double logBase2(double n) {

		return log(n, 2.0);
	}

	/**
	 * Returns the logarithm of a given number to a given base.
	 *
	 * @param n the number
	 * @param base the base
	 * @return the log of the number to the given base
	 */
	public static double log(double n, double base) {

		if (n <= 0) throw new IllegalArgumentException();

		return Math.log(n) / Math.log(base);
	}

	/**
	 * Returns a representation of a byte array as a Hex string.
	 *
	 * @param bytes the byte array
	 * @return a string representation
	 */
	public static String toHexString(byte[] bytes) {

		StringBuffer buffer = new StringBuffer();

		for (byte element : bytes) {

			buffer.append(valueToDigit(element >>> 4 & 0xf, 16));     // 4 most significant bits - shift and mask.
			buffer.append(valueToDigit(element & 0xf, 16));           // 4 least significant bits - mask.
		}
		return buffer.toString();
	}
}
