/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.p2p.interfaces;

import java.math.BigInteger;

/**
 * Interface defining keys.
 *
 * @author sja7, al, stuart, graham
 */
public interface IKey extends Comparable<IKey> {

	/**
	 * @return the key's value
	 */
	public BigInteger keyValue();

	/**
	 * Returns a string representation of the key's value.
	 *
	 * @param radix the radix
	 * @return a string representation of the key value using the given radix
	 */
	public String toString(int radix);

	/**
	 * Returns the number of prefix digits shared with a given key, using a given radix.
	 *
	 * @param key the other key to compare
	 * @param radix the radix to be used
	 * @return the number of shared prefix digits
	 */
	public int countSharedPrefixDigits(IKey key, int radix);

	/**
	 * Returns a given digit of the key, using a given radix.
	 *
	 * @param index the index of the required digit
	 * @param radix the radix to be used
	 * @return the character at that index
	 * @throws IndexOutOfBoundsException if the index does not correspond to a legal digit
	 */
	public char digit(int index, int radix) throws IndexOutOfBoundsException;

	/**
	 * @return a string representing the fraction of the key space represented by this key
	 */
	public String toStringAsKeyspaceFraction();

	/**
	 * @return the number of bits in the minimal two's-complement representation of this key, excluding sign bit
	 */
	public int bitLength();

	/**
	 * @return the number of bits in the minimal two's-complement representation of the biggest key possible
	 */
	public int maxBitLength();

	/**
	 * Calculates the distance from this key to 'end', moving clockwise around the ring.
	 *
	 * @param end a key to which the distance is to be calculated
	 * @return the ring distance between this key and 'end'
	 */
	public BigInteger ringDistanceTo(IKey end);

	/**
	 * Determines whether 'first' is closer to this key than 'second' in absolute terms.
	 *
	 * @param first the first key to be compared
	 * @param second the second key to be compared
	 * @return true if 'first' is closer in ring distance to this key than 'second'
	 */
	public boolean firstNumericallyCloserThanSecond(IKey first, IKey second);

	/**
	 * Determines whether 'first' is closer to this key than 'second', moving clockwise round the ring.
	 *
	 * @param first the first key to be compared
	 * @param second the second key to be compared
	 * @return true if 'first' is closer in ring distance to this key than 'second'
	 */
	public boolean firstCloserInRingThanSecond(IKey first, IKey second);
}
