/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Created on 03-Dec-2004
 */
package uk.ac.standrews.cs.nds.p2p.simulation.impl;

import java.net.InetSocketAddress;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import uk.ac.standrews.cs.nds.p2p.interfaces.IDistanceCalculator;
import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.interfaces.INodeFactory;
import uk.ac.standrews.cs.nds.p2p.interfaces.IP2PNode;
import uk.ac.standrews.cs.nds.p2p.simulation.interfaces.IP2PSimulation;
import uk.ac.standrews.cs.nds.p2p.simulation.interfaces.ISpanTreeGen;
import uk.ac.standrews.cs.nds.p2p.simulation.util.SimProgressWindow;
import uk.ac.standrews.cs.nds.p2p.simulation.util.SimulatedPhysicalDistanceModel;
import uk.ac.standrews.cs.nds.p2p.simulation.util.TopDownSpanTreeGen;
import uk.ac.standrews.cs.nds.p2p.util.SHA1KeyFactory;
import uk.ac.standrews.cs.nds.util.Diagnostic;
import uk.ac.standrews.cs.nds.util.DiagnosticLevel;
import uk.ac.standrews.cs.nds.util.ErrorHandling;


/**
 * @author stuart, al, graham
 * A generic Simulation framework from which specific tests can be constructed.
 * Not specific to any node kind.
 */
public abstract class SimulationFramework<T extends IP2PNode> implements IP2PSimulation<T> {

	private int node_creation_count;

    /******************* Protected Simulation constants *******************/

    protected static final int DEFAULT_NODE_COUNT = 1000;
    protected static final int DEFAULT_NEIGHBOURHOOD_SIZE = 10;
    protected static final int DEFAULT_PROGRESS_GRANULARITY = 1000;
    protected static final int DEFAULT_LINEBREAK_GRANULARITY = 80000;

    protected static final long RANDOM_SEED = 4837275;

    /******************* Protected State *******************/

    protected int progress_granularity = DEFAULT_PROGRESS_GRANULARITY;
    protected int linebreak_granularity = DEFAULT_LINEBREAK_GRANULARITY;
    protected int seed = 43523897;

    protected boolean showProgress; //should a progress bar be displayed

	protected INodeFactory<T> nfactory;		// Factory for constructing nodes
	protected Random rand;				// The random number generator used in simulations
	protected List<T> nodes;			// Holds the live P2P nodes in the network - use distance (difference between index) in this list as real life distance metric
	protected SortedSet<T> nodes_in_key_order; // stores nodes in key order
	protected Map<IKey, Integer> key_to_index; 		//setup hash map from key to index in node array - used to calc distance.
	protected IP2PNode[] neighbours; 		// they made me do this - al
	protected IDistanceCalculator dc;
	protected ISpanTreeGen iSpanTreeGen;
	private final SHA1KeyFactory key_factory;

    /******************* Constructors *******************/
	private SimulationFramework(int node_count, INodeFactory<T> node_factory){
		if (node_factory == null) throw new InvalidParameterException("node factory was null");
		rand = new Random( seed );
		node_creation_count=0;
	    dc = new SimulatedPhysicalDistanceModel(node_count);
	    nfactory = node_factory;
	    key_factory = new SHA1KeyFactory();

	    nodes = new ArrayList<T>();	// Holds nodes
        nodes_in_key_order = new TreeSet<T>(); 	// use this to store nodes in key order
        key_to_index = new HashMap<IKey, Integer>(); 			//setup hash map from key to index in node array
	}

	public SimulationFramework(int node_count, INodeFactory<T> node_factory, boolean showProgress) {
		this(node_count,node_factory);

	    try {
	        SimProgressWindow p=null;
	        if(showProgress)
				p = new SimProgressWindow("Creating nodes",1,node_count);

	        for (int i = 0; i < node_count; i++) {
	        	addNewNode(null);
	            showProgress(i, progress_granularity, linebreak_granularity, DiagnosticLevel.RUN );
	            if(showProgress)
					p.incrementProgress();
	        }

	        if(showProgress)
				p.dispose();
	        iSpanTreeGen = new TopDownSpanTreeGen<T>(this);
	    } catch (Exception e) {
	        e.printStackTrace();
	    }
	}

	public SimulationFramework(String[] keyStrings, INodeFactory<T> nf, boolean showProgress) {
		this(keyStrings.length,nf);
		int node_count=keyStrings.length;

		try {
	        SimProgressWindow p=null;
	        if(showProgress)
				p = new SimProgressWindow("Creating nodes",1,node_count);

	        for (int i = 0; i < keyStrings.length; i++) {
	        	addNewNode(keyStrings[i]);
	            showProgress(i, progress_granularity, linebreak_granularity, DiagnosticLevel.RUN );
	            if(showProgress)
					p.incrementProgress();
	        }

	        if(showProgress)
				p.dispose();
	        iSpanTreeGen = new TopDownSpanTreeGen<T>(this);
	    } catch (Exception e) {
	        e.printStackTrace();
	    }
	}

	protected T addNewNode(String string) {
		IKey k= null;
		int node_number = node_creation_count++;
		String ip = int2IPString(node_number);
		if(string == null)
			k = key_factory.generateKey(ip + node_number);
		else
			k = key_factory.recreateKey(string);
		T newNode = nfactory.makeNode(new InetSocketAddress(ip, node_number),k);
		nodes.add(newNode);
		nodes_in_key_order.add(newNode); 				//add node to the treeset to get them in key order (smallest key first)
		key_to_index.put(k,nodes.indexOf(newNode)); 	//add mapping from key to array index for this node
		return newNode;
	}

	protected boolean removeNode(T node){
		boolean res=nodes.remove(node);
		if(res){
			nodes_in_key_order.remove(node);
			key_to_index.remove(node.getKey());
		}
		return res;
	}

    /******************* Selectors *******************/

    /**
     * @return returns the node_count
     */
    public int getNodeCount() {
        return nodes.size();
    }

    /**
     * @return Returns the hm.
     */
    public Map<IKey, Integer> getKey_to_index() {
        return key_to_index;
    }

    /**
     * @return Returns the nodes.
     */
    public List<T> getNodes() {
        return nodes;
    }

    /**
     * @return returns the nodes_in_key_order
     */
    public SortedSet<T> getNodesInKeyOrder() {
        return nodes_in_key_order;
    }

    public IDistanceCalculator getDistanceCalculator() {
        return dc;
    }

	public ISpanTreeGen getSpanTreeGen() {
		return iSpanTreeGen;
	}
    /******************* Utility Methods *******************/

    protected static void showProgress(int progressCount, int granularity, int lineBreak, DiagnosticLevel level ){
		if(progressCount>0){
			if(progressCount>0 && progressCount % granularity==0) Diagnostic.traceNoSourceNoLn(level,".");
			if(progressCount>0 && progressCount % lineBreak==0) Diagnostic.traceNoSource( level, progressCount + " ");
		}
	}

	/**
	 * @param k
	 * @return Returns the index in the node array of the node with key k. If
	 * no such node exists, -1 is returned.
	 */
	public int key2Index(IKey k) {
		int result = -1;
		Integer index = key_to_index.get(k);
		if (index != null) result = index.intValue();
		return result;
	}

	public static String int2IPString(int ip) { // TODO is this used now we have DC?
		StringBuffer buf = new StringBuffer();
	    buf.append(ip >> 24 & 0xff);
	    buf.append('.');
	    buf.append(ip >> 16 & 0xff);
	    buf.append('.');
	    buf.append(ip >>  8 & 0xff);
	    buf.append('.');
	    buf.append( ip  & 0xff);

	    return new String(buf);
	}

    public void showRoute( T node, IKey k, Route<T> route ) {

        try {
            Diagnostic.trace(DiagnosticLevel.NONE, "index = " + key2Index(node.getKey()) + "\tip= "+ node.getAddress().getAddress() + "\ttarget   " + k + " keylen = " + k.toString().length());
        } catch (Exception e) {
            ErrorHandling.exceptionError(e, "error getting representation of node");
        }

        int hops = 1;

        for (IP2PNode next_node : route)
			try {
                IKey resultKey = next_node.getKey();

                Integer resultindex = key_to_index.get( resultKey );
                Diagnostic.trace( DiagnosticLevel.NONE, "index = " + resultindex + "\tip= "+ node.getAddress().getAddress() + "\tHop    ", + hops++ + " " + resultKey + " keylen = " + resultKey.toString().length());
            }
            catch (Exception e) { ErrorHandling.exceptionError(e, "error getting representation of node"); }
    }

    @Override
	public String toString(){
        return "Simulation nodes: "+nodes.size();
    }

    public void showAll() {
        for(T node:nodes)
			showNode(node);
    }

    /******************* Abstract Methods *******************/

	public abstract void showNode(T node);

	public Random getRand(){
		return rand;
	}
}
