/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.p2p.simulation.util;

import java.util.Enumeration;

import javax.swing.tree.DefaultMutableTreeNode;

import uk.ac.standrews.cs.nds.p2p.interfaces.IP2PNode;
import uk.ac.standrews.cs.nds.p2p.simulation.interfaces.IP2PSimulation;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * This class takes a JCSim object and computes a number of statistics related
 * to the routing trees. For each node in the simulation 3 statistics are
 * computed.
 * <ul>
 * <li>The average physical distance of all routes to the node from each of the
 * nodes in the simulation. That is the total distance for all paths to the node
 * divided by the number of paths to the node (which is just the number of
 * nodes).</li>
 * <li>The average number of routing hops of all routes to the node from each
 * of the nodes in the simulation.</li>
 * <li>The depth of the expected meet points for all routes to a node.</li>
 * </ul>
 *
 * @author stuart
 */

public class AllPathsToAllNodes extends AllPathsToRandomSample {
	private int popularity[];

	/**
	 * @param simulation
	 * @param showProgress
	 */
	public AllPathsToAllNodes(IP2PSimulation<?> simulation, boolean showProgress) {
		super(simulation,showProgress,1.0);
		popularity=null;
	}

	private void updatePopularityStats(DefaultMutableTreeNode tree){
		Enumeration<?> e=tree.breadthFirstEnumeration();
		while(e.hasMoreElements()){
			DefaultMutableTreeNode treenode=(DefaultMutableTreeNode)e.nextElement();
			IP2PNode node=(IP2PNode)treenode.getUserObject();
			try {
				int index=simulation.key2Index(node.getKey());
				popularity[index]+=treenode.getChildCount();
			} catch (Exception e1) {
				ErrorHandling.exceptionError(e1, "ChordNode call threw exception");
				e1.printStackTrace();
			}
		}
	}

	/**
	 * @return Returns the popularity.
	 */
	public int[] getPopularityData() {
		computeStats();
		return popularity;
	}

	@Override
	public void computeStats(){
		if(averageHopCounts==null){
			SimProgressWindow progress=null;
			if(showProgress) {
				progress=new SimProgressWindow("Computing stats for "+node_count+" nodes",0,node_count);
			}
			averageHopCounts=new double[node_count];
			averagePhysicalDistance=new double[node_count];
			depthsOfExpectedMeets=new double[node_count];
			popularity=new int[node_count];
			for(int i=0;i<node_count;i++){
				DefaultMutableTreeNode tree=computePathsToNode(i);
				averageHopCounts[i]=computeAverageHopCountForNode(tree);
				averagePhysicalDistance[i]=computeAveragePhysicalDistanceForNode(tree);
				depthsOfExpectedMeets[i]=computeExpectedMeetDepthForNode(tree);
				updatePopularityStats(tree);
				if(progress!=null) {
					progress.incrementProgress();
				}
			}
			if(progress!=null) {
				progress.dispose();
			}
		}
	}

	/**
	 * @return statistics
	 */
	public String getStats(){
		if(stats==null){
			computeStats();
			stats=simulation.toString();
			stats+="\n\nColumns:\tN.I. (Node Index)\n";
			stats+="\t\tA.P.D. (Average Physical Distance)\n";
			stats+="\t\tA.H.C. (Average Hop Count)\n";
			stats+="\t\tE.M.D. (Expected Meet Depth)\n";
			stats+="\t\tPop. (Popularity)\n";
			stats+="\nN.I.\tA.P.D.\tA.H.C.\tE.M.D.\tPop.\tKey\n";
			for(int i=0;i<averagePhysicalDistance.length;i++) {
				try {
					stats+=i+"\t"+df.format(averagePhysicalDistance[i])
					+"\t"+df.format(averageHopCounts[i])
					+"\t"+df.format(depthsOfExpectedMeets[i])
					+"\t"+df.format(popularity[i])
					+"\t"+simulation.getNodes().get(i).getKey().keyValue().toString(16)
					+"\n";
				} catch (Exception e) {
					ErrorHandling.exceptionError(e, "ChordNode call threw exception");
					e.printStackTrace();
				}
			}
		}
		return stats;
	}
}
