/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.p2p.simulation.util;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import javax.swing.tree.DefaultMutableTreeNode;

import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;
import uk.ac.standrews.cs.nds.p2p.interfaces.IP2PNode;
import uk.ac.standrews.cs.nds.p2p.simulation.impl.Route;
import uk.ac.standrews.cs.nds.p2p.simulation.interfaces.IP2PSimulation;
import uk.ac.standrews.cs.nds.p2p.simulation.interfaces.ISpanTreeGen;
import uk.ac.standrews.cs.nds.p2p.util.SHA1KeyFactory;
import uk.ac.standrews.cs.nds.util.ErrorHandling;

/**
 * @author al
 */
public class TopDownSpanTreeGen<T extends IP2PNode> implements ISpanTreeGen {
    private final IP2PSimulation<T> simulation;

	/**
     * @param simulation
     */
    public TopDownSpanTreeGen(IP2PSimulation<T> simulation) {
       this.simulation=simulation;
    }

    public DefaultMutableTreeNode allPathsToNode(IKey k) {

    	List<T> l = simulation.getNodes();
	    Route<T> rt = simulation.makeRoute(l.get(1),k );    // route from 1 to k.
	    IKey rootKey = null;
        try {
            rootKey = rt.lastHop().getKey();
        } catch (Exception e) {
            ErrorHandling.exceptionError(e, "error getting node representation");
        }
        int rootIndex = simulation.key2Index(rootKey);
	    DefaultMutableTreeNode  tree = new DefaultMutableTreeNode( rt.lastHop() );  // set up a new tree - root is the target of route
	    for( int i = 0 ; i < simulation.getNodeCount(); i++ )
			if(i!=rootIndex) {
				rt = simulation.makeRoute(simulation.getNodes().get(i),k ); // find a route to k from node i
				addRouteToTree( tree,rt );
				//simulation.showRoute(nodes[i],k,rt );
			}
	    return tree;
	}

	public DefaultMutableTreeNode computeAllPathsToRandomNode() {
	    IKey k = new SHA1KeyFactory().generateKey(Integer.toString(simulation.getRand().nextInt()));
	    return allPathsToNode(k);
	}

	protected void addRouteToTree( DefaultMutableTreeNode tree, Route<?> rt ) {
	    try {
	        Iterator<? extends IP2PNode> the_route = rt.reverseIterator();
	        if( tree == null || rt == null )
				ErrorHandling.hardError("Empty root/route passed");
	        if( !the_route.next().getKey().equals( ((IP2PNode)tree.getUserObject()).getKey() ))
				ErrorHandling.hardError("Roots do not match");
	        while( the_route.hasNext() ) {
	            // if we don't have any nodes left in our route we have finished.
	            // We have a next node in the route so get it
	        	IP2PNode next = the_route.next();
	            tree=addNodeToTree(next,tree);
	        }
	        addNodeToTree(rt.getStart(),tree);
	    } catch (Exception e) {
	        e.printStackTrace();
	    }
	}

	protected DefaultMutableTreeNode addNodeToTree(IP2PNode j, DefaultMutableTreeNode tree){
	    IKey nextKey = null;
	    try {
	        nextKey = j.getKey();
	    } catch (Exception e) {
	        ErrorHandling.exceptionError(e, "error getting node representation");
	    }
	    // lookup key in current node and see if Chord node is in tree already
	    DefaultMutableTreeNode child = findChild( tree,nextKey );
	    if( child == null ) { // next node is not in tree so add it as a child of current tree
	        child = new DefaultMutableTreeNode( j );
	        tree.add( child );
	    }
	    return child;
	}

    /**
     * @param tree a DefaultMutableTreeNode containing a partially build tree
     * @param key - the key for which are searching.
     * @return a child of node tree containing a ChordNode with key key
     */
    protected DefaultMutableTreeNode findChild(DefaultMutableTreeNode tree, IKey key) {
        Enumeration<?> children = tree.children();
        while( children.hasMoreElements() ) {
            DefaultMutableTreeNode next = (DefaultMutableTreeNode) children.nextElement();
            IP2PNode nextNode = (IP2PNode) next.getUserObject();
            try {
                if( key.compareTo( nextNode.getKey() ) == 0 )
					return next;
            } catch (Exception e) {
                ErrorHandling.exceptionError(e, "error getting node representation");
            }
        }
        // we haven't found it.
        return null;
    }
}
