/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Created on 26-Oct-2004
 */
package uk.ac.standrews.cs.nds.p2p.util;

import java.net.InetSocketAddress;

import uk.ac.standrews.cs.nds.p2p.impl.Key;
import uk.ac.standrews.cs.nds.p2p.interfaces.IHash;
import uk.ac.standrews.cs.nds.p2p.interfaces.IKey;

/**
 * Provides various ways to generate keys.
 *
 * @author stuart, graham
 */
public class HashBasedKeyFactory {

	private final IHash hash;
	private final RandomGenerator random_generator;

	// ***********************************************************************

	/**
	 * Creates a key factory using a given hashing function.
	 *
	 * @param hash the hash function
	 */
	public HashBasedKeyFactory(IHash hash) {

		this.hash = hash;
		random_generator = new RandomGenerator(hash);
	}

	// ***********************************************************************

	/**
	 * Creates a key with a value generated by hashing the given byte array.
	 *
	 * @param bytes the array from which to generate the key's value
	 * @return a key with a value generated from bytes
	 */
	public IKey generateKey(byte[] bytes) {

		byte[] hashed = hash.hash(bytes);
		return new Key(hashed);
	}

	/**
	 * Creates a key with a value generated by hashing the given string.
	 *
	 * @param s the string from which to generate the key's value
	 * @return a key with a value generated from s
	 */
	public IKey generateKey(String s) {

		return generateKey(s.getBytes());
	}

	/**
	 * Creates a key with a value generated by hashing the given IP address.
	 *
	 * @param ip_address the IP address from which to generate the key's value
	 * @return a key with a value generated from ip_address
	 */
	public IKey generateKey(InetSocketAddress ip_address) {

		String s = ip_address.getAddress().getHostAddress() + ip_address.getPort();
		return generateKey(s);
	}

	/**
	 * Creates a key with a value represented by the given string. The representation is in the same
	 * format as obtained by calling toString() on a key.
	 *
	 * @param s the representation of a key value
	 * @return a key with the given value
	 */
	public IKey recreateKey(String s) {

		return new Key(s);
	}

	/**
	 * Creates a key with a pseudo-random value. The method {@link #setDeterministic(boolean)} controls whether the sequence
	 * generated by successive calls is deterministic.
	 *
	 * @return a key with a pseudo-random value
	 */
	public IKey generateRandomKey() {

		byte[] bytes = random_generator.generateRandomBytes(hash.getHashSize());

		return generateKey(bytes);
	}

	/**
	 * Controls whether the sequence of values generated by successive calls to {@link #generateRandomKey()} is deterministic.
	 *
	 * @param deterministic true if the sequence should be deterministic
	 */
	public void setDeterministic(boolean deterministic) {

		random_generator.setDeterministic(deterministic);
	}
}
