/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * Created on Aug 8, 2006 at 5:02:53 PM.
 */
package uk.ac.standrews.cs.nds.util;

import java.io.IOException;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashSet;
import java.util.Set;

import com.mindbright.ssh2.SSH2ConsoleRemote;
import com.mindbright.ssh2.SSH2Exception;

/**
 * Allows files/modules to be checked out/updated on a remote host (say A) from a remote CVS server (say B). Requirements:
 *
 * 1) SSH service is available on both A and B.
 * 2) A will accept SSH connections from this host, and B will accept SSH connections from A.
 * 3) The shell available on A understands bash syntax.
 *
 * The various methods establish an SSH connection to A and issue bash commands that perform the appropriate remote CVS operations on B.
 *
 * @author graham
 */
public class CVSUtil {

	private static final String PASSPHRASE_PROMPT = "Enter passphrase";

	// These strings must all be the same length.
	private static final String CVS_ERROR_STEM1 =              "Enter passphrase";
	private static final String CVS_ERROR_STEM2 =              "Permission denie";
	private static final String CVS_RESPONSE_FINISHED_MARKER = "XXXXXXXXXXXXXXXX";

	private static final String SSH_WRAPPER_PATH = "/tmp/cvs_ssh_wrapper";

	private static CVSUtil instance = new CVSUtil();

	// ***************************************************************************************

	/**
	 * Checks out a CVS module to a remote host A from a remote CVS server B. Either the password or the passphrase must be non-null. If both are
	 * supplied then A will try to use public key authentication first in connecting to B, failing over to password.
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine to which the module will be checked out
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_module the name of the CVS module to be checked out
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A into which the module will be checked out
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if the connection to the CVS server fails, or the module cannot be checked out
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static synchronized void checkoutCVSModule(SSH2ConnectionWrapper ssh_client_wrapper, InetAddress cvs_server_address, String cvs_username, String cvs_password, String cvs_passphrase, String cvs_module, String cvs_root, String destination_directory_path, boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_server_address, cvs_username, cvs_password, cvs_passphrase, cvs_root, destination_directory_path, "checkout", cvs_module, null, show_cvs_output);
	}

	/**
	 * Updates a CVS module on a remote host A from a remote CVS server B. Either the password or the passphrase must be non-null. If both are
	 * supplied then A will try to use public key authentication first in connecting to B, failing over to password.
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine on which the module will be updated
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_module the name of the CVS module to be updated
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A in which the module will be updated
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if the connection to the CVS server fails, or the module cannot be updated
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static synchronized void updateCVSModule(SSH2ConnectionWrapper ssh_client_wrapper, InetAddress cvs_server_address, String cvs_username, String cvs_password, String cvs_passphrase, String cvs_module, String cvs_root, String destination_directory_path, boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_server_address, cvs_username, cvs_password, cvs_passphrase, cvs_root, destination_directory_path, "update", cvs_module, null, show_cvs_output);
	}

	/**
	 * Checks out a set of specified items within a CVS module to a remote host A from a remote CVS server B. Either the password or the passphrase must be non-null. If both are
	 * supplied then A will try to use public key authentication first in connecting to B, failing over to password.
	 *
	 * Each item is specified as path within the module e.g. "uk/ac/stand/dcs/nds/eventModel/Event.java"
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine to which the module will be checked out
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_module the name of the CVS module to be checked out
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A into which the module will be checked out
	 * @param item_paths a set of paths to items within the CVS module (see note above)
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if the set of items is empty, the connection to the CVS server fails, or the items cannot be checked out
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static synchronized void checkoutCVSItems(SSH2ConnectionWrapper ssh_client_wrapper, InetAddress cvs_server_address, String cvs_username, String cvs_password, String cvs_passphrase, String cvs_module, String cvs_root, String destination_directory_path, Set<String> item_paths, boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		// CVS doesn't allow checking out specific items for the first time (as opposed to updating them, which is allowed). Therefore this checks out the entire module
		// and then deletes everything but the specific items (and the CVS directories, which are needed in case of subsequent updates).

		if (item_paths.size() == 0) throw new CVSException("empty item set");

		String delete_unwanted_items_command = genDeleteUnwantedItems(cvs_module, item_paths);

		invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_server_address, cvs_username, cvs_password, cvs_passphrase, cvs_root, destination_directory_path, "checkout", cvs_module, delete_unwanted_items_command, show_cvs_output);
	}

	/**
	 * Updates a set of specified items within a CVS module on a remote host A from a remote CVS server B. Either the password or the passphrase must be non-null. If both are
	 * supplied then A will try to use public key authentication first in connecting to B, failing over to password.
	 *
	 * Each item is specified as path within the module e.g. "uk/ac/stand/dcs/nds/eventModel/Event.java"
	 *
	 * If 'checkout_if_necessary' is set to true, the current versions of the items will be installed on A regardless of whether the
	 * module had been checked out previously. Otherwise, the update will fail if the items do not already exist on A.
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine to which the module will be checked out
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_module the name of the CVS module to be updated
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A into which the module will be checked out
	 * @param item_paths a set of paths to items within the CVS module (see note above)
	 * @param checkout_if_necessary true if the items should be checked out first if they do not already exist on A (see note above)
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if the set of items is empty, the connection to the CVS server fails, or the items cannot be checked out
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static synchronized void updateCVSItems(SSH2ConnectionWrapper ssh_client_wrapper, InetAddress cvs_server_address, String cvs_username, String cvs_password, String cvs_passphrase, String cvs_module, String cvs_root, String destination_directory_path, Set<String> item_paths, boolean checkout_if_necessary, boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		if (item_paths.size() == 0) throw new CVSException("empty item set");

		// Build the list of items to be updated.
		String update_items_list = "";
		for (String item_path : item_paths) update_items_list += cvs_module + "/" + item_path + " ";

		// Build the command to set the CVS authentication mechanism.
		String cvs_rsh_setup = genAuth(cvs_password, cvs_passphrase);

		if (checkout_if_necessary) {
			// Update the specific items if the module is already present, otherwise check out whole module and delete the rest, as in checkoutCVSItems().

			String remote_update =   genChangeToDestinationDirectoryAndInvokeCVSOperation(cvs_server_address, cvs_username, cvs_root, destination_directory_path, "update", update_items_list, "echo '" + CVS_RESPONSE_FINISHED_MARKER + "'", cvs_rsh_setup);
			String remote_checkout = genChangeToDestinationDirectoryAndInvokeCVSOperation(cvs_server_address, cvs_username, cvs_root, destination_directory_path, "checkout", cvs_module, genDeleteUnwantedItems(cvs_module, item_paths), cvs_rsh_setup);

			String test_whether_installation_directory_exists = "test -d " + destination_directory_path + "/" + cvs_module;

			String remote_command = genRemoteTest(test_whether_installation_directory_exists, remote_update, remote_checkout);

			invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_password, cvs_passphrase, remote_command, show_cvs_output);
		}
		else {
			// Update the specific items regardless of whether the module is present.

			String remote_command = genChangeToDestinationDirectoryAndInvokeCVSOperation(cvs_server_address, cvs_username, cvs_root, destination_directory_path, "update", update_items_list, null, cvs_rsh_setup);

			invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_password, cvs_passphrase, remote_command, show_cvs_output);
		}
	}

	// ***************************************************************************************

	/**
	 * Remotely invokes a specified CVS operation, followed by an optional additional shell command, and interprets the output appropriately.
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A into which the module will be checked out
	 * @param cvs_operation the operation to be invoked, eg "update", "checkout" etc
	 * @param cvs_target the target of the CVS operation, eg a module name or list of files
	 * @param subsequent_command a shell command to be performed after the CVS operation, or null for none
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if authentication with the CVS server fails
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	private static void invokeCVSOperationAndInterpretOutput(SSH2ConnectionWrapper ssh_client_wrapper, InetAddress cvs_server_address, String cvs_username, String cvs_password, String cvs_passphrase, String cvs_root, String destination_directory_path, String cvs_operation, String cvs_target, String subsequent_command, boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		String cvs_rsh_setup = genAuth(cvs_password, cvs_passphrase);

		String remote_command = genChangeToDestinationDirectoryAndInvokeCVSOperation(cvs_server_address, cvs_username, cvs_root, destination_directory_path, cvs_operation, cvs_target, subsequent_command, cvs_rsh_setup);

		invokeCVSOperationAndInterpretOutput(ssh_client_wrapper, cvs_password, cvs_passphrase, remote_command, show_cvs_output);
	}

	/**
	 * Remotely invokes a specified command, consisting of a CVS operation followed by an optional additional shell command, and interprets the output appropriately.
	 *
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 * @param remote_command the command to be executed: CVS operation plus optional additional command
	 * @param show_cvs_output true if the output from the CVS server should be sent to standard output
	 *
	 * @throws CVSException if authentication with the CVS server fails
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	private static void invokeCVSOperationAndInterpretOutput(SSH2ConnectionWrapper ssh_client_wrapper, String cvs_password, String cvs_passphrase, String remote_command, final boolean show_cvs_output) throws CVSException, SSH2Exception, IOException {

		// Used for communication between the asynchronous output processor and the main thread.
		final BooleanWrapper use_passphrase = instance.new BooleanWrapper();
		final BooleanWrapper cvs_error =      instance.new BooleanWrapper();       // Horrible hack required by Java.

		// Used for synchronisation between the asynchronous output processor and the main thread.
		//TODO stu - change these to use java.util.concurrent.Semaphore
		final Semaphore prompt_sync = new Semaphore(0);
		final Semaphore output_sync = new Semaphore(0);

		// Defines how output read from the remote process is handled.
		IStreamProcessor process_remote_output = new IStreamProcessor() {

			String prompt = "";
			String output = "";

			// Keeps track of which processing phase has been reached.
			// Sequence: reading enough of authentication prompt to know which method to use -> flushing remainder of authentication prompt -> reading output from CVS operation -> finished.
			OutputProcessingPhase phase = OutputProcessingPhase.READING_PROMPT;

			public boolean processByte(int byte_value) {

				if (show_cvs_output) System.out.print((char)byte_value);

				switch (phase) {

					case READING_PROMPT : {

						// Still don't know which authentication method is being used.

						prompt += (char)byte_value;

						// Look for '@', which indicates a password prompt, or a specific prompt for a passphrase.

						if (byte_value == '@' || prompt.startsWith(PASSPHRASE_PROMPT)) {

							use_passphrase.value = byte_value != '@';
							prompt_sync.semSignal();
							phase = OutputProcessingPhase.READING_PAST_PROMPT;
						}
						break;
					}

					case READING_PAST_PROMPT : {

						// Now know which authentication method is being used; discard further output until end of line.

						if (byte_value == '\n') phase = OutputProcessingPhase.READING_CVS_OUTPUT;
						break;
					}

					case READING_CVS_OUTPUT : {

						// Still don't know whether CVS operation was successful.

						// Possible responses:
						// 1. Successful CVS operation - output starts "cvs" - actually this just means that a connection with the CVS process was established, but seems good enough for now.
						// 2. Invalid passphrase -       output starts "Enter passphrase"
						// 3. Invalid password -         output starts "Permission denied"

						output += (char)byte_value;

						if (output.length() == CVS_ERROR_STEM1.length()) {

							cvs_error.value = output.equals(CVS_ERROR_STEM1) || output.equals(CVS_ERROR_STEM2);
							output_sync.semSignal();
							phase = OutputProcessingPhase.FINISHED;
						}
						break;
					}

					case FINISHED : {

						// Ignore any further output.
					}
				}

				return true;
			}
		};

		SSH2ConsoleRemote console = Processes.runProcess(remote_command, ssh_client_wrapper, null, null, process_remote_output, null);

		// Suspend until the processing thread discovers which authentication method is being used.
		prompt_sync.semWait();

		// Write the password or passphrase, as appropriate, to the remote process.
		PrintStream remote_input_stream = new PrintStream(console.getStdIn());

		if (use_passphrase.value) remote_input_stream.println(cvs_passphrase);
		else                      remote_input_stream.println(cvs_password);

		remote_input_stream.flush();

		// Suspend until the processing thread decides whether the operation has been successful.
		output_sync.semWait();

		if (cvs_error.value) throw new CVSException("invalid password or passphrase");

		// Tidy up.
		console.waitForExitStatus();
		console.close();
		ssh_client_wrapper.getSSH2Client().getTransport().normalDisconnect("CVS operation completed");
	}

	/**
	 * Generates if-then-else in bash syntax.
	 *
	 * @param condition the condition
	 * @param command_if_true command to be executed if true
	 * @param command_if_false command to be executed if false
	 *
	 * @return the generated command
	 */
	private static String genRemoteTest(String condition, String command_if_true, String command_if_false) {

	    return "if " + condition + "; then " + command_if_true + "; else " + command_if_false + " fi";
    }

	/**
	 * Generates a bash command to change to a given directory, invoke a specified CVS operation, followed by an optional additional bash command.
	 *
	 * @param cvs_server_address the IP address of the CVS server
	 * @param cvs_username the username to be used in connecting to the CVS server
	 * @param cvs_root the path of the CVS repository on B
	 * @param destination_directory_path the path of the directory on A into which the module will be checked out
	 * @param cvs_operation the operation to be invoked, eg "update", "checkout" etc
	 * @param cvs_target the target of the CVS operation, eg a module name or list of files
	 * @param subsequent_command a shell command to be performed after the CVS operation, or null for none
	 * @param cvs_rsh_setup the command to set the CVS_RSH shell variable
	 *
	 * @return the generated command
	 */
	private static String genChangeToDestinationDirectoryAndInvokeCVSOperation(InetAddress cvs_server_address, String cvs_username, String cvs_root, String destination_directory_path, String cvs_operation, String cvs_target, String subsequent_command, String cvs_rsh_setup) {

		StringBuffer command = new StringBuffer("cd " + destination_directory_path + "; " + cvs_rsh_setup);

		command.append("; cvs -d:ext:" + cvs_username + "@" + cvs_server_address.getHostName() + ":" + cvs_root + " " + cvs_operation + " " + cvs_target);

		if (subsequent_command != null) command.append("; " + subsequent_command);

		return command.toString();
	}

	/**
	 * Generates a bash command to set the CVS_RSH shell variable. If a password is supplied but not a passphrase, the ssh connection to the CVS server will be forced to use
	 * password authentication by setting the PreferredAuthentications=password flag. Vice-versa for a passphrase without a password. If both are supplied then it is up to ssh.
	 * Normally that will try public key first and then password if that doesn't work.
	 *
	 * @param cvs_password the password to be used in connecting to the CVS server (optional, may be null)
	 * @param cvs_passphrase the private key passphrase to be used in connecting to the CVS server (optional, may be null)
	 *
	 * @return the generated command
	 * @throws CVSException if the password and passphrase are both null.
	 */
	private static String genAuth(String cvs_password, String cvs_passphrase) throws CVSException {

		// Treat empty string and null the same.
		if (cvs_password.equals(""))   cvs_password = null;
		if (cvs_passphrase.equals("")) cvs_passphrase = null;

		if (cvs_password == null && cvs_passphrase == null) throw new CVSException("neither password nor passphrase specified");

		if (cvs_password == null)   return genAuth("publickey");
		if (cvs_passphrase == null) return genAuth("password");

		return "export CVS_RSH=ssh";
	}

	/**
	 * Generates a bash command to set the CVS_RSH shell variable. It doesn't work setting the shell variable to a call to ssh with the appropriate arguments, so instead
	 * the call with the arguments is written to a temporary file, and the shell variable set to execute that file.
	 *
	 * @param preferred_authentication_method either "publickey" or "password"
	 * @return the generated command
	 */
	private static String genAuth(String preferred_authentication_method) {

		return "echo 'ssh -o PreferredAuthentications=" + preferred_authentication_method + " $*' > " + SSH_WRAPPER_PATH + "; chmod +x " + SSH_WRAPPER_PATH + "; export CVS_RSH=" + SSH_WRAPPER_PATH;
	}

	/**
	 * Generates a bash command to delete everything in the specified checked out module apart from the the specified items and the CVS directories.
	 *
	 * @param cvs_module the name of the CVS module being checked out
	 * @param item_paths a set of paths to items within the CVS module (see note above)
	 *
	 * @return the generated command
	 * @throws CVSException if an item does not conform to URI syntax
	 */
	private static String genDeleteUnwantedItems(String cvs_module, Set<String> item_paths) throws CVSException {

		/*

		This deletes everything but the specific items (and the CVS directories, which are needed in case of subsequent updates).

		Example shell script to be generated for module 'NDS' and items 'uk/ac/stand/dcs/nds/eventModel/Event.java', 'uk/ac/stand/dcs/nds/eventModel/IEvent.java':

		# For each item...

		# Make corresponding directory in /tmp.
		mkdir -p /tmp/NDS/uk/ac/stand/dcs/nds/eventModel;

		# Copy the CVS directories at each level.
		cp -rf NDS/CVS /tmp/NDS
		cp -rf NDS/uk/CVS /tmp/NDS/uk
		cp -rf NDS/uk/ac/CVS /tmp/NDS/uk/ac
		cp -rf NDS/uk/ac/stand/CVS /tmp/NDS/uk/ac/stand
		cp -rf NDS/uk/ac/stand/dcs/CVS /tmp/NDS/uk/ac/stand/dcs
		cp -rf NDS/uk/ac/stand/dcs/nds/CVS /tmp/NDS/uk/ac/stand/dcs/nds
		cp -rf NDS/uk/ac/stand/dcs/nds/eventModel/CVS /tmp/NDS/uk/ac/stand/dcs/nds/eventModel

		# Copy the item.
		cp NDS/uk/ac/stand/dcs/nds/eventModel/Event.java /tmp/NDS/uk/ac/stand/dcs/nds/eventModel

		# ... and repeat for next item.

		cp NDS/uk/ac/stand/dcs/nds/eventModel/IEvent.java /tmp/NDS/uk/ac/stand/dcs/nds/eventModel

		# Now delete everything.
		rm -rf NDS/*

		# And bring back saved stuff from /tmp.
		mkdir -p NDS/uk/ac/stand/dcs/nds/eventModel
		cp -rf /tmp/NDS/CVS NDS
		cp -rf /tmp/NDS/uk/CVS NDS/uk
		cp -rf /tmp/NDS/uk/ac/CVS NDS/uk/ac
		cp -rf /tmp/NDS/uk/ac/stand/CVS NDS/uk/ac/stand
		cp -rf /tmp/NDS/uk/ac/stand/dcs/CVS NDS/uk/ac/stand/dcs
		cp -rf /tmp/NDS/uk/ac/stand/dcs/nds/CVS NDS/uk/ac/stand/dcs/nds
		cp -rf /tmp/NDS/uk/ac/stand/dcs/nds/eventModel/CVS NDS/uk/ac/stand/dcs/nds/eventModel

		cp /tmp/NDS/uk/ac/stand/dcs/nds/eventModel/Event.java NDS/uk/ac/stand/dcs/nds/eventModel

		cp /tmp/NDS/uk/ac/stand/dcs/nds/eventModel/IEvent.java NDS/uk/ac/stand/dcs/nds/eventModel

		*/

		StringBuffer move_wanted_items_command =    new StringBuffer();
		StringBuffer restore_wanted_items_command = new StringBuffer();

		Set<String> encountered_directories = new HashSet<String>();

		try {
			for (String item_path : item_paths) {

				URI uri = new URI("/" + item_path);

				String item_parent = UriUtil.parentUri(uri).toString();

				// Check whether parent has already been encountered.
				if (!encountered_directories.contains(item_parent)) {

					move_wanted_items_command.append("mkdir -p /tmp/" + cvs_module + item_parent + "; ");
					move_wanted_items_command.append("cp -rf " + cvs_module + "/CVS /tmp/NDS; ");

					restore_wanted_items_command.append("mkdir -p " + cvs_module + item_parent + "; ");
					restore_wanted_items_command.append("cp -rf /tmp/" + cvs_module + "/CVS " + cvs_module + "; ");

					String intermediate_path = "";

					for (String path_element : UriUtil.pathElementIterator(new URI(item_parent))) {

						intermediate_path += path_element;

						move_wanted_items_command.append(   "cp -rf " + cvs_module + "/" + intermediate_path + "/CVS /tmp/" + cvs_module + "/" + intermediate_path + "; ");
						restore_wanted_items_command.append("cp -rf /tmp/" + cvs_module + "/" + intermediate_path + "/CVS " + cvs_module + "/" + intermediate_path + "; ");

						intermediate_path += "/";
					}

					encountered_directories.add(item_parent);
				}

				move_wanted_items_command.append(   "cp " + cvs_module + "/" + item_path + " /tmp/" + cvs_module + item_parent + "; ");
				restore_wanted_items_command.append("cp /tmp/" + cvs_module + "/" + item_path + " " + cvs_module + item_parent + "; ");
			}
		}
		catch (URISyntaxException e) { throw new CVSException("invalid item path syntax"); }

		String delete_all_items_command = "rm -rf " + cvs_module + "/*; ";

		return move_wanted_items_command + delete_all_items_command + restore_wanted_items_command;
	}

	class BooleanWrapper {
		protected boolean value = false;
	}

	enum OutputProcessingPhase {

		/**
		 *
		 */
		READING_PROMPT,

		/**
		 *
		 */
		READING_PAST_PROMPT,

		/**
		 *
		 */
		READING_CVS_OUTPUT,

		/**
		 *
		 */
		FINISHED;
	}
}
