/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

/**
 * Representation of a Java class path.
 *
 * @author graham
 */
public class ClassPath implements Iterable<File> {

	private static final String LOCAL_PATH_SEPARATOR = getPathSeparator();

	private List<File> classpath_entries;

	// ***********************************************************************

	/**
	 * @return the class path of the currently executing program
	 */
	public static ClassPath getCurrentClassPath() {

		// Work around bug on Windows/Eclipse where part of the classpath contains forward slashes.
		String environment_class_path = System.getProperty("java.class.path").replace("/", SystemUtil.LOCAL_FILE_SEPARATOR);

		return new ClassPath(environment_class_path);
	}

	// ***********************************************************************

	/**
	 * Creates a new empty class path.
	 */
	public ClassPath() {

		classpath_entries = new ArrayList<File>();
	}

	/**
	 * Creates a new class path with entries specified in standard string notation for this platform.
	 *
	 * @param classpath_string the class path entries
	 */
	public ClassPath(String classpath_string) {

		this(classpath_string.split(LOCAL_PATH_SEPARATOR));
	}

	/**
	 * Creates a new class path with specified entries.
	 *
	 * @param classpath_element_paths the class path entries
	 */
	public ClassPath(String[] classpath_element_paths) {

		this();

		for (String classpath_element_path : classpath_element_paths)
			classpath_entries.add(new File(classpath_element_path));
	}

	/**
	 * Creates a new class path with specified entries.
	 *
	 * @param classpath_elements the class path entries
	 */
	public ClassPath(File[] classpath_elements) {

		this();
		classpath_entries.addAll(Arrays.asList(classpath_elements));
	}

	/**
	 * Creates a new class path with specified entries.
	 *
	 * @param classpath_elements the class path entries
	 */
	public ClassPath(Iterable<File> classpath_elements) {

		this();
		for (File entry : classpath_elements) classpath_entries.add(entry);
	}

	// ***********************************************************************

	/**
	 * Returns a representation of the class path formatted for the current local platform.
	 *
	 * @param use_path_quote true if the resulting string should be surrounded with appropriate quote characters
	 * @return a representation of the class path
	 */
	public synchronized String toString(boolean use_path_quote) {

		return toString(SystemUtil.LOCAL_PLATFORM, use_path_quote);
	}

	/**
	 * Returns a representation of the class path formatted for the given platform.
	 *
	 * @param platform a platform
	 * @return a representation of the class path
	 */
	public synchronized String toString(PlatformDescriptor platform) {

		return toString(platform, true);
	}

	/**
	 * Returns a representation of the class path formatted for the given platform.
	 *
	 * @param platform a platform
	 * @param use_path_quote true if the resulting string should be surrounded with appropriate quote characters
	 * @return a representation of the class path
	 */
	public synchronized String toString(PlatformDescriptor platform, boolean use_path_quote) {

		String path_separator = getPathSeparator(platform);
		String path_quote =     use_path_quote ? SystemUtil.getPathQuote(platform) : "";

		StringBuffer classpath = new StringBuffer(path_quote);

		boolean first = true;
		for (File path_element : classpath_entries) {

			if (first) first = false;
			else       classpath.append(path_separator);

			classpath.append(path_element.getAbsolutePath());
		}
		classpath.append(path_quote);

		return classpath.toString();
	}

	/**
	 * Adds an entry to the class path at the given position. Position 0 corresponds to the start of the class path.
	 *
	 * @param index the index at which the entry should be added
	 * @param entry the entry to be added
	 */
	public void add(int index, File entry) {

		classpath_entries.add(index, entry);
	}

	/**
	 * Adds an entry to the end of the class path.
	 *
	 * @param entry the entry to be added
	 */
	public void append(File entry) {

		classpath_entries.add(entry);
	}

	/**
	 * Adds all entries in the given class path to this class path.
	 *
	 * @param other_path the other class path
	 */
	public void append(ClassPath other_path) {

		classpath_entries.addAll(other_path.classpath_entries);
	}

	// ***********************************************************************

	/**
	 * @return an iterator over the entries in the class path, each an instance of File representing either a directory or a jar file.
	 */
	public Iterator<File> iterator() {

		return classpath_entries.iterator();
	}

	// ***********************************************************************

	/**
	 * Modifies all relative paths to resolve them relative to the given root.
	 *
	 * @param root a root directory
	 */
	public void resolveRelativePaths(File root) {

		List<File> new_classpath_entries = new ArrayList<File>();

		for (File entry : classpath_entries)
			if (entry.isAbsolute()) new_classpath_entries.add(entry);
			else new_classpath_entries.add(
					new File(root.getAbsolutePath() +
					File.separator + entry.getPath()));

		classpath_entries = new_classpath_entries;
	}

	@Override
	public ClassPath clone() {

		return new ClassPath(classpath_entries);
	}

	@Override
	public String toString() {

		return toString(SystemUtil.LOCAL_PLATFORM);
	}

	// ***********************************************************************

	private static String getPathSeparator() {

		return System.getProperty("path.separator");
	}

	private static String getPathSeparator(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_MAC :     return ":";
			case PLATFORM_LINUX :   return ":";
			case PLATFORM_WINDOWS : return ";";

			default : ErrorHandling.hardError("unknown platform"); return null;
		}
	}
}
