/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 *
 */
package uk.ac.standrews.cs.nds.util;

/**
 * A global threshold diagnostic level may be set by the user.
 * The ordering of levels is FULL < INIT < RUN < RUNALL < RESULT < FINAL < NONE.
 *
 * @author al, graham
 */
public enum DiagnosticLevel {

	/**
	 * Level 0.
	 */
	FULL,

	/**
	 * Level 1.
	 */
	INIT,

	/**
	 * Level 2.
	 */
	RUN,

	/**
	 * Level 3.
	 */
	RUNALL,

	/**
	 * Level 4.
	 */
	RESULT,

	/**
	 * Level 5.
	 */
	FINAL,

	/**
	 * Level 6.
	 */
	NONE;

	private static final int FULL_VALUE =   0;
	private static final int INIT_VALUE =   1;
	private static final int RUN_VALUE =    2;
	private static final int RUNALL_VALUE = 3;
	private static final int RESULT_VALUE = 4;
	private static final int FINAL_VALUE =  5;
	private static final int NONE_VALUE =   6;

	/**
	 * @return the numerically lowest diagnostic level.
	 */
	public static DiagnosticLevel getMin() {
		return FULL;
	}

	/**
	 * @return the numerically highest diagnostic level.
	 */
	public static DiagnosticLevel getMax() {
		return NONE;
	}

	/**
	 * @return the value of the numerically lowest diagnostic level.
	 */
	public static int getMinValue() {
		return getMin().numericalValue();
	}

	/**
	 * @return the value of the numerically highest diagnostic level.
	 */
	public static int getMaxValue() {
		return getMax().numericalValue();
	}

	/**
	 * Converts a numerical value to a diagnostic level.
	 *
	 * @param value a numerical value
	 * @return the corresponding level
	 */
	public static DiagnosticLevel fromNumericalValue(int value) {

		switch (value) {

			case FULL_VALUE :   return FULL;
			case INIT_VALUE :   return INIT;
			case RUN_VALUE :    return RUN;
			case RUNALL_VALUE : return RUNALL;
			case RESULT_VALUE : return RESULT;
			case FINAL_VALUE :  return FINAL;
			case NONE_VALUE :   return NONE;
		}

		return null;
	}

	/**
	 * @return the numerical value of this diagnostic level
	 */
	public int numericalValue() {

		switch (this) {

			case FULL :   return FULL_VALUE;
			case INIT :   return INIT_VALUE;
			case RUN :    return RUN_VALUE;
			case RUNALL : return RUNALL_VALUE;
			case RESULT : return RESULT_VALUE;
			case FINAL :  return FINAL_VALUE;
			case NONE :   return NONE_VALUE;
		}

		return -1;
	}

	/**
	 * Tests whether this diagnostic level meets a given threshold.
	 *
	 * @param threshold a threshold level
	 * @return true if this diagnostic level is numerically greater than or equal to the value of the given level
	 */
	public boolean meetsThreshold(DiagnosticLevel threshold) {

		return numericalValue() >= threshold.numericalValue();
	}
}
