/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.mail.MessagingException;
import javax.mail.internet.MimeUtility;

/**
 * Provides utility methods for encoding and decoding byte arrays.
 *
 * @author graham
 */
public class Encoding {

	/**
	 * Translates the given byte array to a string using the specified encoding.
	 *
	 * @param bytes a byte array
	 * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
	 * @return a string representing the encoding of the given bytes
	 * @throws MessagingException if the encoding is not recognised
	 * @throws IOException if an error occurs during translation
	 */
	public static String encodeByteArray(byte[] bytes, String encoding) throws MessagingException, IOException {

		ByteArrayOutputStream encoded_bytes_stream = new ByteArrayOutputStream();
		OutputStream encoding_stream = MimeUtility.encode(encoded_bytes_stream, encoding);

		encoding_stream.write(bytes);
		encoding_stream.close();

		return new String(encoded_bytes_stream.toByteArray());
	}

	/**
	 * Translates the given byte array to a string using the specified encoding, with the option of suppressing certain characters.
	 * Note that there is no facility to reverse an encoding if character suppression is used.
	 *
	 * @param bytes a byte array
	 * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
	 * @param suppressed_characters a string containing characters that should be suppressed from the output
	 * @return a string representing the encoding of the given bytes
	 * @throws MessagingException if the encoding is not recognised
	 * @throws IOException if an error occurs during translation
	 */
	public static String encodeByteArray(byte[] bytes, String encoding, String suppressed_characters) throws MessagingException, IOException {

		String encoded = encodeByteArray(bytes, encoding);
		StringBuffer buffer = new StringBuffer();

		int len = encoded.length();
		for (int i = 0; i < len; i++) {
			char c = encoded.charAt(i);
			if (suppressed_characters.indexOf(c) == -1) buffer.append(c);
		}

		return buffer.toString();
	}

	/**
	 * Translates the given encoded string to a decoded byte array using the specified encoding.
	 *
	 * @param text an encoded string
	 * @param result_size the size of the decoded array to be returned
	 * @param encoding the encoding as defined at http://java.sun.com/javaee/5/docs/api/javax/mail/internet/MimeUtility.html
	 * @return an array representing the decoding of the given string
	 * @throws MessagingException if the encoding is not recognised
	 * @throws IOException if an error occurs during translation
	 */
	public static byte[] decodeByteArray(String text, int result_size, String encoding) throws MessagingException, IOException {

		byte[] encoded_bytes = text.getBytes();
		InputStream decoding_stream = MimeUtility.decode(new ByteArrayInputStream(encoded_bytes), encoding);

	 	byte[] decoded_bytes = new byte[result_size];
		decoding_stream.read(decoded_bytes, 0, result_size);

		return decoded_bytes;
	}

	/**
	 * Converts a byte array to a hex representation.
	 *
	 * This method derived from code in:
	 *
	 * SHA1.java - An implementation of the SHA-1 Algorithm
	 * obtained from
	 * http://jabberapplet.jabberstudio.org/doxygen/SHA1_8java-source.html
	 *
	 * This version by Chuck McManis (cmcmanis@netcom.com) and
	 * still public domain.
	 *
	 * Adapted for JabberApplet by M.Kiesel
	 *
	 * Based on the C code that Steve Reid wrote his header
	 * was :
	 *      SHA-1 in C
	 *      By Steve Reid <steve@edmweb.com>
	 *      100% Public Domain
	 *
	 * @param bytes a byte array
	 * @return a string representation of the bytes in hex
	 */
	public static String byteArrayToHex(byte[] bytes) {

		StringBuffer sb = new StringBuffer();

		for (byte digestBit : bytes) {

			char c1 = (char)(digestBit >>> 4 & 0xf);
			char c2 = (char)(digestBit & 0xf);

			c1 = (char)(c1 > 9 ? 'a' + (c1 - 10) : '0' + c1);
			c2 = (char)(c2 > 9 ? 'a' + (c2 - 10) : '0' + c2);

			sb.append(c1);
			sb.append(c2);
		}

		return sb.toString();
	}
}
