/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.util;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import uk.ac.standrews.cs.nds.eventModel.EventFactory;
import uk.ac.standrews.cs.nds.eventModel.eventBus.busInterfaces.IEventBus;

/**
 * Provides support for various error message output.
 *
 * Based on code from the com.findnearyou package (C)
 * A. Dearle & R.Connor.
 *
 * @author al, graham
 */
public class ErrorHandling {

	private static final ErrorHandling instance = new ErrorHandling();
	private static final String SEPARATOR = " : ";
	private static final String HARD_ERROR_LABEL = "FATAL ERROR: ";

	private static boolean use_timestamp = false;
	private static boolean local_reporting = true;

	private IEventBus bus;

	protected static Object sync = new Object();    // To allow Diagnostic calls to be synchronised with respect to these methods.

	protected static final String DEFAULT_DATE_FORMAT_PATTERN = "HHmm.ss.SSS yyyy-MM-dd";
	private static DateFormat dateformat = new SimpleDateFormat(DEFAULT_DATE_FORMAT_PATTERN);

	//********************************************************************************************************************************************/

	public static void setTimestampFlag(boolean use_timestamp) {

		ErrorHandling.use_timestamp = use_timestamp;
	}

	public static DateFormat getDateFormat(){
		return dateformat;
	}
	/**
	 * Outputs an error message on the event bus and to standard error, if enabled.
	 *
	 * @param msg a descriptive message
	 */
	public static void error(Object... msg) {
		outputError(true, false, Diagnostic.getMethodInCallChain(), null, msg);
	}

	/**
	 * Outputs an error message on the event bus and to standard error, if enabled.
	 *
	 * @param msg a descriptive message
	 */
	public static void errorNoSource(Object... msg) {
		outputError(true, false, null, null, msg);
	}

	/**
	 * Outputs an error message to standard error, without sending on the event bus.
	 *
	 * @param msg a descriptive message
	 */
	public static void errorNoEvent(Object... msg) {
		outputError(false, false, Diagnostic.getMethodInCallChain(), null, msg);
	}

	/**
	 * Outputs an error message to standard error regardless of whether local
	 * reporting is enabled or not. The message is displayed without displaying
	 * the source of the error, without the "ErrorHandling:" prefix and without sending
	 * on the event bus.
	 *
	 * @param msg a descriptive message
	 */
	public static void errorExplicitLocalReport(Object... msg) {
		outputError(false, false, null, null, msg);
	}

	/**
	 * Outputs an error message to standard error regardless of whether local
	 * reporting is enabled or not, and performs a system exit. The message is
	 * displayed without displaying the source of the error, without the
	 * "ErrorHandling:" prefix and without sending on the event bus.
	 *
	 * @param msg a descriptive message
	 */
	public static void hardErrorExplicitLocalReport(Object... msg) {
		outputError(false, true, null, null, msg);
		hardExit();
	}

	/********************************************************************************************************************************************/

	/**
	 * Outputs an error message, and performs a system exit.
	 *
	 * @param msg a descriptive message
	 */
	public static void hardError(Object... msg) {
		outputError(true, true, Diagnostic.getMethodInCallChain(), null, msg);
		hardExit();
	}

	/**
	 * Outputs an error message, and performs a system exit.
	 *
	 * @param msg a descriptive message
	 */
	public static void hardErrorNoSource(Object... msg) {
		outputError(true, true, null, null, msg);
		hardExit();
	}

	/**
	 * Outputs an error message to standard error, without sending on the event bus, and performs a system exit.
	 *
	 * @param msg a descriptive message
	 */
	public static void hardErrorNoEvent(Object... msg) {
		outputError(false, true, Diagnostic.getMethodInCallChain(), null, msg);
		hardExit();
	}

	/********************************************************************************************************************************************/

	/**
	 * Outputs details of an exception, followed by a stack trace.
	 * @param e the exception
	 * @param msg a descriptive message
	 */
	public static void exceptionError(Throwable e, Object... msg) {
		outputError(true, false, Diagnostic.getMethodInCallChain(), e, msg);
	}

	/**
	 * Outputs details of an exception, followed by a stack trace, without sending on the event bus.
	 * @param e the exception
	 * @param msg a descriptive message
	 */
	public static void exceptionErrorNoEvent(Throwable e, Object... msg) {
		outputError(false,false,Diagnostic.getMethodInCallChain(),e, msg);
	}

	/**
	 * Outputs details of an exception, followed by a stack trace, and performs a system exit.
	 * @param e the exception
	 * @param msg a descriptive message
	 */
	public static void hardExceptionError(Throwable e, Object... msg) {
		exceptionError(e, msg);
		hardExit();
	}

	/**
	 * Outputs details of an exception, followed by a stack trace, without sending on the event bus, and performs a system exit.
	 * @param e the exception
	 * @param msg a descriptive message
	 */
	public static void hardExceptionErrorNoEvent(Throwable e, Object... msg) {
		exceptionErrorNoEvent(e, msg);
		hardExit();
	}

	//********************************************************************************************************************************************/

	public static synchronized void initialise(IEventBus bus) {
		if (instance.bus == null)
			instance.bus = bus;
		else {
			outputError(true,false, "ErrorHandling::initialise", null, "Attempt to reinitialise ErrorHandling singleton");
			hardExit();
		}
	}

	public static void enableLocalErrorReporting() {

		local_reporting = true;
	}

	public static void disableLocalErrorReporting() {

		local_reporting = false;
	}

	/********************************************************************************************************************************************/

	private ErrorHandling() {

		bus = null;
	}

	private static void hardExit() {
		System.exit(-1);
	}

	//********************************************************************************************************************************************/

	private static void outputError(boolean generate_event, boolean fatal, String source, Throwable e, Object... messages) {

		synchronized (sync) {

			StringBuffer sb = new StringBuffer();
			if (fatal) sb.append(HARD_ERROR_LABEL + " ");

			if(use_timestamp){
				//add current time to the error message
				sb.append("[");
				Date now = new Date();
				String formatted_date = dateformat.format(now);
				sb.append(formatted_date);
				sb.append("] ");
			}

			if (source != null && source.length() > 0) {
				sb.append(source);
				sb.append(SEPARATOR);
			}

			for (Object message : messages)
				/*if (message!= null)*/ sb.append(message);

			if (e != null) {
				sb.append(" Error: ");
				sb.append(e.getMessage());
				sb.append("\n");
				sb.append(getStackTrace(e));
			}

			String buffer_as_string = sb.toString();

			if (generate_event && instance.bus != null)
				instance.bus.publishEvent(EventFactory.makeErrorEvent(buffer_as_string));

			if (local_reporting)
				System.err.println(buffer_as_string);
		}
	}

	private static String getStackTrace(Throwable e) {

		StringBuffer sb = new StringBuffer();

		sb.append(e);
		sb.append("\n");
		StackTraceElement[] trace = e.getStackTrace();
		for (StackTraceElement element : trace)
			sb.append("\tat " + element + "\n");

		return sb.toString();
	}
}