/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.util;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.Enumeration;

/**
 * Various network utilities.
 *
 * @author stuart
 */
public class NetworkUtil {

	/**
	 * The undefined port.
	 */
	public static final int UNDEFINED_PORT = -1;

	/**
	 * Returns the first non-loopback IPv4 address (Inet4Address) that can be found for an interface on the local host.
	 * This method should be used in place of InetAddress.getLocalHost(), which may return an Inet6Address object
	 * corresponding to the IPv6 address of a local interface. The bind operation is not supported by this
	 * address family.
	 *
	 * @return the first IPv4 address found
	 * @throws UnknownHostException if no IPv4 address can be found
	 */
	public static InetAddress getLocalIPv4Address() throws UnknownHostException {

		InetAddress default_local_address = InetAddress.getLocalHost();

		// Return the default local address if it's an IPv4 address and isn't the loopback address. This will work in most cases.
		if (default_local_address instanceof Inet4Address && !default_local_address.isLoopbackAddress())
			return default_local_address;

		// Otherwise, look for an IPv4 address among the other interfaces.
		InetAddress loopback_address = null;

		try {
			Enumeration<NetworkInterface> interfaces_enumeration = NetworkInterface.getNetworkInterfaces();

			if (interfaces_enumeration != null) {
				for (NetworkInterface network_interface : Collections.list(interfaces_enumeration)) {
					for (InetAddress address : Collections.list(network_interface.getInetAddresses()))

						if (address instanceof Inet4Address)
							// Found an IPv4 address. Return it if it's not loopback, otherwise remember it.
							if (!address.isLoopbackAddress()) return address;
							else {
								loopback_address = address;
							}
				}
			}

			// Haven't found any non-loopback IPv4 address, so return loopback if available.
			if (loopback_address != null) return loopback_address;
		}
		catch (SocketException e) {
			// Ignore.
		}

		throw new UnknownHostException("local host has no interface with an IPv4 address");
	}

	
	
	/**
	 * Returns an InetSocketAddress corresponding to a local non-loopback IPv4 address.
	 *
	 * @param port the port
	 * @return the corresponding local address
	 *
	 * @throws IllegalArgumentException if the port is outwith the legal range
	 * @throws UnknownHostException if no IPv4 address can be found
	 */
	public static InetSocketAddress getLocalIPv4InetSocketAddress(int port) throws IllegalArgumentException, UnknownHostException {

		return new InetSocketAddress(getLocalIPv4Address(), port);
	}

	/**
	 * Outputs a diagnostic trace containing the address of a given host.
	 *
	 * @param prefix a trace prefix
	 * @param host the host
	 */
	public static void reportHostAddress(String prefix, InetSocketAddress host) {

		Diagnostic.trace(DiagnosticLevel.INIT, prefix, ": ", formatHostAddress(host));
	}

	/**
	 * Extracts an InetSocketAddress from a string of the form "host:port". If the host part is empty, the local
	 * loopback address is used. If the port part is empty, the specified default port is used.
	 *
	 * @param host_and_port a string of the form "host:port"
	 * @param default_port the default port to be used if the port is not specified
	 * @return a corresponding InetSocketAddress
	 *
	 * @throws UnknownHostException if the specified host cannot be resolved
	 * @throws NumberFormatException if the port number cannot be parsed
	 */
	public static InetSocketAddress processHostPortParameter(String host_and_port, int default_port) throws UnknownHostException, NumberFormatException {

		String host_name = extractHostName(host_and_port);
		int port =         extractPortNumber(host_and_port);

		if (host_name == "") {
			if (port == UNDEFINED_PORT) return getLocalIPv4InetSocketAddress(default_port);
			else                        return getLocalIPv4InetSocketAddress(port);
		}
		else if (port == UNDEFINED_PORT) return new InetSocketAddress(InetAddress.getByName(host_name), default_port);
		else                             return new InetSocketAddress(InetAddress.getByName(host_name), port);
	}

	/**
	 * Extracts a host name from a string of the form "[host][:][port]". If the
	 * host part is empty, the empty string is returned.
	 *
	 * @param host_and_port a string of the form "host:port", "host", "host:", ":port"
	 * @return the host name
	 */
	public static String extractHostName(String host_and_port) {

		if (host_and_port == null) return "";

		int separator_index = host_and_port.indexOf(":");

		if (separator_index != -1) return host_and_port.substring(0, separator_index);
		else                       return host_and_port;                                // No port was specified.
	}

	/**
	 * Extracts a port number as a string from a string of the form "[host][:][port]". If
	 * the port part is empty, the string representation of {@link #UNDEFINED_PORT} is returned.
	 *
	 * @param host_and_port a string of the form "host:port", "host", "host:" or ":port"
	 * @return the port number as a string
	 */
	public static String extractPortNumberAsString(String host_and_port) {

		if (host_and_port == null) return String.valueOf(UNDEFINED_PORT);

		int separator_index = host_and_port.indexOf(":");

		// Check for "<host>", "<host>:" and ":"
		if (separator_index == -1 || separator_index == host_and_port.length() - 1) return String.valueOf(UNDEFINED_PORT);

		return host_and_port.substring(separator_index + 1);
	}

	/**
	 * Extracts a port number from a string of the form "[host][:][port]". If
	 * the port part is empty {@link #UNDEFINED_PORT} is returned.
	 *
	 * @param host_and_port a string of the form "host:port", "host", "host:" or ":port"
	 * @return the port number
	 */
	public static int extractPortNumber(String host_and_port) {

		try {
			return Integer.parseInt(extractPortNumberAsString(host_and_port));
		}
		catch (NumberFormatException e) { return UNDEFINED_PORT; }
	}

	/**
	 * Tests whether a given address is a valid local address.
	 *
	 * @param address an address
	 * @return true if the address is a valid local address
	 */
	public static boolean isValidLocalAddress(InetAddress address) {

		try {
			for (NetworkInterface network_interface : Collections.list(NetworkInterface.getNetworkInterfaces())) {
				for (InetAddress network_address : Collections.list(network_interface.getInetAddresses()))
					if (address.equals(network_address)) return true;
			}
		}
		catch (SocketException e) {
			// Ignore.
		}

		return false;
	}

	/**
	 * Returns a description of a given host address.
	 *
	 * @param address an address
	 * @return a description of the address
	 */
	public static String formatHostAddress(InetSocketAddress address) {

		if (address != null) {

			String host = address.getAddress().getHostAddress();
			int port =    address.getPort();

			return formatHostAddress(host,port);
		}
		else return null;
	}

	/**
	 * Returns a description of a given host address.
	 *
	 * @param host an IP address
	 * @param port a port
	 * @return a description of the address
	 */
	public static String formatHostAddress(String host, int port) {

		return host + ":" + port;
	}

	/**
	 * Returns a description of a given socket.
	 *
	 * @param socket a socket
	 * @return a description of the socket's address
	 */
	public static String formatHostAddress(Socket socket) {

		return formatHostAddress(socket.getInetAddress(), socket.getPort());
	}

	/**
	 * Returns a description of a given host address.
	 *
	 * @param address an address
	 * @param port a port
	 * @return a description of the address
	 */
	public static String formatHostAddress(InetAddress address, int port) {

		return formatHostAddress(address.getHostAddress(), port);
	}
}
