/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Created on Jan 26, 2005 at 11:49:28 PM.
 */
package uk.ac.standrews.cs.nds.util;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import com.mindbright.ssh2.SSH2ConsoleRemote;
import com.mindbright.ssh2.SSH2Exception;

/**
 * Utility class for creating local or remote OS processes.
 *
 * @author graham
 */
public class Processes {

	/**
	 * Shell command to echo current PID.
	 */
	private static final String ECHO_PID_COMMAND = "echo $$ ";

	/**
	 * Shell command to destroy a process.
	 */
	private static final String KILL_COMMAND = "kill -9 ";

	/**
	 * Shell command to get the children of a process.
	 */
	private static final String PSTREE_COMMAND = "pstree -p ";

	/**
	 * Separator character in shell command.
	 */
	private static final String SHELL_SEPARATOR = "; ";

	/**
	 * Used to help interpret output of process echoing its PID.
	 */
	protected static final String MARKER1 = "XXX";
	protected static final String MARKER2 = "YYY";

	// ***********************************************************************

	/**
	 * Runs the specified command in a new OS process. Output is written to local System.out and System.err.
	 *
	 * @param command the command to be executed
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 */
	public static Process runProcess(String command) throws IOException {

		return runProcess(command, System.out, System.err);
	}

	/**
	 * Runs the specified command in a new OS process.
	 *
	 * @param command the command to be executed
	 * @param output_stream the stream to which output from the process should be directed
	 * @param error_stream the stream to which error output from the process should be directed
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 */
	public static Process runProcess(String command, final OutputStream output_stream, final OutputStream error_stream) throws IOException {

		// Run the command in a new OS shell.
		Process p = Runtime.getRuntime().exec(command);

		final PrintStream output_print_stream = new PrintStream(output_stream);
		final PrintStream error_print_stream =  new PrintStream(error_stream);

		// Get streams reading from the standard out and standard error for the new process.
		final BufferedReader command_output_stream = new BufferedReader(new InputStreamReader(new BufferedInputStream(p.getInputStream())));
		final BufferedReader command_error_stream =  new BufferedReader(new InputStreamReader(new BufferedInputStream(p.getErrorStream())));

		// Start a thread to capture any output and write it locally.
		new Thread() {

			@Override
			public void run() {

				String s = null;
				try {
					while ((s = command_output_stream.readLine()) != null) output_print_stream.println(s);
					command_output_stream.close();

				} catch (IOException e) { /* Ignore read errors; they mean process is done. */ } // $codepro.audit.disable emptyCatchClause

			}
		}.start();

		// Start a thread to capture any errors and write them locally.
		new Thread() {

			@Override
			public void run() {

				String s = null;
				try {
					while ((s = command_error_stream.readLine()) != null) error_print_stream.println(s);
					command_error_stream.close();

				} catch (IOException e) { /* Ignore read errors; they mean process is done. */ } // $codepro.audit.disable emptyCatchClause
			}
		}.start();

		return p;
	}

	// ***********************************************************************

	/**
	 * Runs the specified class in a new local OS process, using the class path of the currently executing program. Output is written to local System.out and System.err.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args) throws IOException {

		String java_command = SystemUtil.getJavaCommand(clazz, args);
		return runProcess(java_command);
	}

	/**
	 * Runs the specified class in a new local OS process, using the specified class path. Output is written to local System.out and System.err.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @param class_path the class path
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args, ClassPath class_path) throws IOException {

		String java_command = SystemUtil.getJavaCommand(clazz, args, class_path);
		return runProcess(java_command);
	}

	// ***********************************************************************

	/**
	 * Runs the specified class in a new remote OS process, using the class path of the currently executing program. Output is written to local System.out and System.err.
	 * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path.
	 *
	 * @param clazz the class to be executed
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runJavaProcess(Class<?> clazz, SSH2ConnectionWrapper ssh_client_wrapper) throws IOException, SSH2Exception {

		PlatformDescriptor platform = PlatformDescriptor.getPlatform(ssh_client_wrapper);
		List<String> args = new ArrayList<String>();
		String java_command = SystemUtil.getJavaCommand(clazz, args, platform);

		return runProcess(java_command, ssh_client_wrapper);
	}

	/**
	 * Runs the specified class in a new remote OS process, using the class path of the currently executing program. Output is written to local System.out and System.err.
	 * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args, SSH2ConnectionWrapper ssh_client_wrapper) throws IOException, SSH2Exception {

		PlatformDescriptor platform = PlatformDescriptor.getPlatform(ssh_client_wrapper);
		String java_command = SystemUtil.getJavaCommand(clazz, args, platform);

		return runProcess(java_command, ssh_client_wrapper);
	}

	/**
	 * Runs the specified class in a new remote OS process, using the specified class path. Output is written to local System.out and System.err.
	 * It is assumed that the necessary classes are already installed remotely at the locations specified in the class path.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param class_path the class path
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args, SSH2ConnectionWrapper ssh_client_wrapper, ClassPath class_path) throws IOException, SSH2Exception {

		PlatformDescriptor platform = PlatformDescriptor.getPlatform(ssh_client_wrapper);
		String java_command = SystemUtil.getJavaCommand(clazz, args, class_path, platform);

		return runProcess(java_command, ssh_client_wrapper);
	}

	/**
	 * Runs the specified class in a new remote OS process. The class path is inferred, if possible, from the class's corresponding project
	 * meta-data and the specified code base location. Output is written to local System.out and System.err.
	 * It is assumed that the necessary classes are already installed remotely at the specified code base location.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param code_base the path of the remote code base
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args, SSH2ConnectionWrapper ssh_client_wrapper, File code_base) throws IOException, SSH2Exception {

		throw new UnsupportedOperationException();
	}

	/**
	 * Runs the specified class in a new remote OS process. The class path is inferred, if possible, from the class's corresponding project
	 * meta-data and the specified code base location. Output is written to local System.out and System.err.
	 * If required, the necessary classes are installed remotely at the specified code base location.
	 *
	 * @param clazz the class to be executed
	 * @param args command line arguments
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param code_base the path of the remote code base
	 * @param code_server the address of an http server from which the necessary classes can be retrieved
	 * @param always_install true if the classes should be re-installed even if already present
	 *
	 * @return a handle to the new OS process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runJavaProcess(Class<?> clazz, List<String> args, SSH2ConnectionWrapper ssh_client_wrapper, File code_base, URL code_server, boolean always_install) throws IOException, SSH2Exception {

		throw new UnsupportedOperationException();
	}

	// ***********************************************************************

	/**
	 * Runs the specified command on a remote machine accessed via SSH. Output from the remote process is passed to standard out and standard error.
	 *
	 * @param command the command to be executed
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 *
	 * @return a handle to the remote process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runProcess(String command, SSH2ConnectionWrapper ssh_client_wrapper) throws SSH2Exception, IOException {

		return runProcess(command, ssh_client_wrapper, System.out, System.err);
	}

	/**
	 * Runs the specified command on a remote machine accessed via SSH.
	 *
	 * @param command the command to be executed
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param output_stream the stream to which output from the process should be directed
	 * @param error_stream the stream to which error output from the process should be directed
	 *
	 * @return a handle to the remote process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static Process runProcess(String command, final SSH2ConnectionWrapper ssh_client_wrapper, final OutputStream output_stream, final OutputStream error_stream) throws SSH2Exception, IOException {

		/*
		 * The MindTerm package supports the creation of a remote process, but not the later destruction of that process should it become necessary. So:
		 *
		 * 1. Prefix the remote user command with a shell command to echo the current process ID (PID).
		 * 2. Execute the combined commands and process the standard output to extract the PID.
		 * 3. That gives the PID of the parent shell process.
		 * 4. Now execute a remote 'pstree' command to get the children of the parent process.
		 * 5. Process the output of that to get the PID of the user process.
		 * 6. Store that PID so it can be used in a remote 'kill' command if the process needs to be destroyed.
		 */

		Diagnostic.trace(DiagnosticLevel.RUN, "Remote command: ", command);

		// Prefix the command with code to print the process ID to the output stream.
		// This will be the ID of the shell command not of the user command itself.
		final String modified_command = ECHO_PID_COMMAND + MARKER1 + SHELL_SEPARATOR + command;

		// Object to contain the PID once found.
		final ProcessID process_id = new Processes().new ProcessID();

		// Processor to read the output from the remote process and extract the parent PID.
		IStreamProcessor extract_parent_PID = new IStreamProcessor() {

			/**
			 * Used to accumulate output from the remote process while scanning for PID information.
			 */
			String main_output = "";

			/**
			 * Indicates when processing has been completed.
			 */
			boolean parent_scan_done = false;

			public boolean processByte(int byte_value) throws IOException {

				// If processing is finished then just pass the byte on to the stream.
				if (parent_scan_done) return true;

				// Accumulate the output without passing it on.
				main_output += (char)byte_value;

				// Test whether the marker indicating the end of the PID has been reached.
				if (main_output.endsWith(MARKER1)) {

					// Assume that the PID starts at the beginning of the stream.
					String parent_PID = main_output.substring(0, main_output.length() - MARKER1.length());
					try {
						getChildPID(parent_PID);
					}
					catch (SSH2Exception e) { throw new IOException(e.getMessage()); }
					parent_scan_done = true;
				}

				return false;
			}

			public void setChildPID(String pid) {

				process_id.child_PID = pid;
				process_id.semaphore.semSignal();
			}

			public void getChildPID(String pid) throws SSH2Exception, IOException {

				// Command to output the children of the given process.
				final String get_child_command = PSTREE_COMMAND + pid + "; echo " + MARKER2;

				// Processor to read the output from the remote process and extract the child PID.
				IStreamProcessor extract_child_PID = new IStreamProcessor() {

					/**
					 * Used to accumulate output from the remote process while scanning for child PID information.
					 */
					String pstree_output = "";

					/**
					 * Indicates when processing has been completed.
					 */
					boolean child_scan_done = false;

					public boolean processByte(int byte_value) {

						if (!child_scan_done) {

							// Accumulate the output.
							pstree_output += (char)byte_value;

							if (pstree_output.endsWith(MARKER2)) {

								setChildPID(extractChildPIDFromStream(pstree_output));

								child_scan_done = true;
							}
						}

						return false;
					}

					private String extractChildPIDFromStream(String stream_output) {

						// Expected output from remote process on Linux:
						// bash(PID1)---<child process name>(PID2)
						//
						// Expected output from remote process on Mac:
						// \--- PID2 <owner> <child process name>
						//
						// This is the format returned by the Darwin Ports version of pstree <http://pstree.darwinports.com/>
						//
						// So look first for the \--- pattern indicating Mac. If not found, assume Linux and read back from the end to find an opening bracket.

						String mac_pattern = "\\---";

						int pattern_start = stream_output.indexOf(mac_pattern);
						if (pattern_start > -1) {

							// Assume Mac.
							while (stream_output.indexOf(mac_pattern, pattern_start + 1) > -1) pattern_start = stream_output.indexOf(mac_pattern, pattern_start + 1);

							int first_pid_char_index = pattern_start + 5;
							int last_pid_char_index = first_pid_char_index;
							while (last_pid_char_index < stream_output.length() && stream_output.charAt(last_pid_char_index) != ' ') last_pid_char_index++;
							return stream_output.substring(first_pid_char_index, last_pid_char_index);
						}
						else {

							// Assume Linux.
							int i = stream_output.length() - 1;

							// Read back to find the corresponding opening bracket.
							while (i >= 0 && stream_output.charAt(i) != '(') i--;

							if (i >= 0) return stream_output.substring(i + 1, stream_output.length() - MARKER2.length() - 2);
							else return "";
						}
					}
				};

				runProcess(get_child_command, ssh_client_wrapper, null, null, extract_child_PID, null);
			}
		};

		// Run the combined command to echo the parent PID and run the user command.
		final SSH2ConsoleRemote console = runProcess(modified_command, ssh_client_wrapper, output_stream, error_stream, extract_parent_PID, null);

		// Wait until the PID is available.
		process_id.semaphore.semWait();

		// Return a Java process handle that allows the remote process to be killed.
		return new Process() {

			@Override
			public void destroy() {

				String kill_command = KILL_COMMAND + process_id.child_PID;
				try {
					runProcess(kill_command, ssh_client_wrapper, null, null, null, null);
				}
				catch (Exception e) {
					Diagnostic.trace(DiagnosticLevel.RUN, "exception while trying to kill remote process: ", e.getMessage());
				}

				console.close();
			}

			@Override
			public int exitValue() {
				return waitFor();
			}

			@Override
			public InputStream getErrorStream() {
				return null;
			}

			@Override
			public InputStream getInputStream() {
				return console.getStdOut();
			}

			@Override
			public OutputStream getOutputStream() {
				return console.getStdIn();
			}

			@Override
			public int waitFor() {
				return console.waitForExitStatus();
			}
		};
	}

	/**
	 * Runs the specified command on a remote machine accessed via SSH, using public key authentication. Output from the remote process is
	 * passed through the specified processors, and then directed to the specified streams.
	 *
	 * Based on example code from MindTerm (http://www.mindterm.com/).
	 *
	 * @param command the command to be executed
	 * @param ssh_client_wrapper an SSH connection wrapper for a remote machine
	 * @param output_stream the stream to which output from the process should be directed
	 * @param error_stream the stream to which error output from the process should be directed
	 * @param output_processor a processor for the standard output from the new process
	 * @param error_processor a processor for the error output from the new process
	 *
	 * @return a handle to the remote process
	 * @throws IOException if an error occurs when reading output from the process
	 * @throws SSH2Exception if an SSH connection to the remote machine cannot be established
	 */
	public static SSH2ConsoleRemote runProcess(final String command, final SSH2ConnectionWrapper ssh_client_wrapper, OutputStream output_stream, OutputStream error_stream,
			IStreamProcessor output_processor, IStreamProcessor error_processor) throws SSH2Exception, IOException {

		/*
		 * Create the remote console to use for command execution.
		 */
		final SSH2ConsoleRemote console = new SSH2ConsoleRemote(ssh_client_wrapper.getSSH2Client().getConnection());

		final OutputStream intercept_output_stream = makeInterceptStream(output_stream, output_processor);
		final OutputStream intercept_error_stream =  makeInterceptStream(error_stream, error_processor);

		// Create a thread to run the command remotely via SSH.
		new Thread() {

			@Override
			public void run() {

				// Run the command (returns a boolean indicating success, ignored here).
				console.command(command, intercept_output_stream, intercept_error_stream, true);  // Set TTY.

				// Wait for the remote command to complete or be killed.
				console.waitForExitStatus();
			}

		}.start();

		return console;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Creates an output stream that processes bytes using the specified processor and passes further output to the specified stream.
	 *
	 * @param stream OutputStream
	 * @param processor IStreamProcessor
	 * @return OutputStream
	 */
	public static OutputStream makeInterceptStream(final OutputStream stream, final IStreamProcessor processor) {

		return new OutputStream() {

			@Override
			public void write(int b) throws IOException {

				// Allow further processing if the processor says so, or if there is no processor.
				boolean further_processing = processor == null || processor.processByte(b);

				if (further_processing && stream != null) stream.write(b);
			}

			@Override
			public void close() throws IOException {

				// Do nothing.
			}

			@Override
			public void flush() throws IOException {

				stream.flush();
			}
		};
	}

	/**
	 * Used to store a PID for a remote process.
	 */
	private class ProcessID {

		/**
		 * The PID of the remote process.
		 */
		String child_PID = "";

		/**
		 * Used to synchronise access to the PID value.
		 */
		//TODO stu - change this to use java.util.concurrent.Semaphore
		Semaphore semaphore = new Semaphore(0);
	}
}
