/*
 *  NDS Library
 *  Copyright (C) 2006-2008 Distributed Systems Architecture Research Group
 *  http://www-systems.cs.st-andrews.ac.uk/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.ac.standrews.cs.nds.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;

/**
 * Provides various utilities to do with platform specific aspects.
 *
 * @author graham
 */
public class SystemUtil {

	private static final String TOOLS_JAR_NAME = "tools.jar";
	private static final String JDK_DIR_PREFIX = "jdk";
	
	static final PlatformDescriptor LOCAL_PLATFORM = PlatformDescriptor.getPlatform();
	static final String LOCAL_FILE_SEPARATOR = getFileSeparator();

	// ***********************************************************************

	/**
	 * @return the file separator character for the current platform
	 */
	public static String getFileSeparator() {

		return System.getProperty("file.separator");
	}

	/**
	 * @param platform a platform descriptor
	 * @return the file separator character for the specified platform
	 */
	public static String getFileSeparator(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_MAC :     return "/";
			case PLATFORM_LINUX :   return "/";
			case PLATFORM_WINDOWS : return "\\";

			default : ErrorHandling.hardError("unknown platform"); return null;
		}
	}

	// ***********************************************************************

	/**
	 * Gets the path of the local Java SDK.
	 *
	 * @return the path of the local Java SDK
	 * @throws FileNotFoundException if the path does not exist locally
	 */
	public static File getJavaSDKPath() throws FileNotFoundException {

		return getJavaSDKPath(LOCAL_PLATFORM, true);
	}

	/**
	 * Gets the path of the Java SDK on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @return the expected path of the Java SDK on the specified platform
	 */
	public static File getJavaSDKPath(PlatformDescriptor platform) {

		try {
			return getJavaSDKPath(platform, false);
		}
		catch (FileNotFoundException e) { throw new RuntimeException("unexpected file not found exception: shouldn't be checked"); }
	}

	/**
	 * Gets the path of the Java SDK on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @param check true if the local presence of the SDK should be checked
	 * @return the expected path of the Java SDK on the specified platform
	 * @throws FileNotFoundException if check is true and the path does not exist locally
	 */
	public static File getJavaSDKPath(PlatformDescriptor platform, boolean check) throws FileNotFoundException {

		return getJavaSDKPath(platform, System.getProperty("java.version"), check);
	}

	/**
	 * Gets the path of the Java SDK on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @param java_version the Java version on the specified platform
	 * @param check true if the local presence of the SDK should be checked
	 * @return the expected path of the Java SDK on the specified platform
	 * @throws FileNotFoundException if check is true and the path does not exist locally
	 */
	public static File getJavaSDKPath(PlatformDescriptor platform, String java_version, boolean check) throws FileNotFoundException {

		// For Mac the SDK has a fixed alias as defined in properties file.
		// For Linux and Windows, the parent directory is fixed, as defined in properties file.

		File sdk_path = null;

		switch (platform) {
			case PLATFORM_LINUX : {
				sdk_path = new File(StringConstants.getString("SystemUtil.DEFAULT_JAVA_SDK_PATH_PARENT_LINUX") + getFileSeparator(platform) + JDK_DIR_PREFIX + java_version);
				break;
			}
			case PLATFORM_WINDOWS : {
				sdk_path = new File(StringConstants.getString("SystemUtil.DEFAULT_JAVA_SDK_PATH_PARENT_WINDOWS") + getFileSeparator(platform) + JDK_DIR_PREFIX + java_version);
				break;
			}
			case PLATFORM_MAC : {
				sdk_path = new File(StringConstants.getString("SystemUtil.DEFAULT_JAVA_SDK_PATH_MAC"));
				break;
			}

			default : { ErrorHandling.hardError("unknown platform: " + platform); }
		}

		if (check && !sdk_path.exists()) throw new FileNotFoundException("SDK path not found: " + sdk_path.getAbsolutePath());

		return sdk_path;
	}

	// ***********************************************************************

	/**
	 * Gets the path of the Java compiler executable.
	 *
	 * @return the path of the local Java compiler
	 * @throws FileNotFoundException if the path does not exist locally
	 */
	public static File getJavaCompilerPath() throws FileNotFoundException {

		return getJavaCompilerPath(LOCAL_PLATFORM);
	}

	/**
	 * Gets the path of the Java compiler executable on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @return the expected path of the Java compiler executable on the specified platform
	 */
	public static File getJavaCompilerPath(PlatformDescriptor platform) {

		try {
			return getJavaCompilerPath(platform, false);
		}
		catch (FileNotFoundException e) { throw new RuntimeException("unexpected file not found exception: shouldn't be checked"); }
	}

	/**
	 * Gets the path of the Java compiler executable on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @param check true if the local presence of the compiler should be checked
	 * @return the expected path of the Java compiler executable on the specified platform
	 * @throws FileNotFoundException if check is true and the path does not exist locally
	 */
	public static File getJavaCompilerPath(PlatformDescriptor platform, boolean check) throws FileNotFoundException {

		File compiler_path = new File(new File(getJavaSDKPath(), getJavaExecutableDirectoryName(LOCAL_PLATFORM)), getJavaCompilerName());

		if (check && !compiler_path.exists()) throw new FileNotFoundException("compiler path not found: " + compiler_path.getAbsolutePath());

		return compiler_path;
	}

	// ***********************************************************************

	/**
	 * Gets the path of the Java interpreter executable.
	 *
	 * @return the path of the local Java interpreter
	 * @throws FileNotFoundException if the path does not exist locally
	 */
	public static File getJavaInterpreterPath() throws FileNotFoundException {

		return getJavaInterpreterPath(LOCAL_PLATFORM);
	}

	/**
	 * Gets the path of the Java interpreter executable on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @return the expected path of the Java interpreter executable on the specified platform
	 */
	public static File getJavaInterpreterPath(PlatformDescriptor platform) {
//TODO fix 
		//try {
			String DEFAULT_JAVA_SDK_PATH="C:\\Program Files\\Java\\jdk1.5.0_12\\bin\\java.exe";
			return new File(DEFAULT_JAVA_SDK_PATH);
			//return getJavaInterpreterPath(platform, false);
		//}
		//catch (FileNotFoundException e) { throw new RuntimeException("unexpected file not found exception: shouldn't be checked"); }
	}

	/**
	 * Gets the path of the Java interpreter executable on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @param check true if the local presence of the interpreter should be checked
	 * @return the expected path of the Java interpreter executable on the specified platform
	 * @throws FileNotFoundException if check is true and the path does not exist locally
	 */
	public static File getJavaInterpreterPath(PlatformDescriptor platform, boolean check) throws FileNotFoundException {
		
		

		File interpreter_path = new File(new File(getJavaSDKPath(), getJavaExecutableDirectoryName(platform)), getJavaInterpreterName());

		if (check && !interpreter_path.exists()) throw new FileNotFoundException("interpreter path not found: " + interpreter_path.getAbsolutePath());

		return interpreter_path;
	}

	// ***********************************************************************

	/**
	 * Gets the path of the tools.jar file.
	 *
	 * @return the path of the local tools.jar file, or null if it is not needed
	 * @throws FileNotFoundException if the file is needed and it does not exist locally
	 */
	public static File getJavaToolsJarPath() throws FileNotFoundException {

		return getJavaToolsJarPath(LOCAL_PLATFORM);
	}

	/**
	 * Gets the path of the tools.jar file on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @return the expected path of the tools.jar file on the specified platform, or null if it is not needed
	 */
	public static File getJavaToolsJarPath(PlatformDescriptor platform) {

		try {
			return getJavaToolsJarPath(platform, false);
		}
		catch (FileNotFoundException e) { throw new RuntimeException("unexpected file not found exception: shouldn't be checked"); }
	}

	/**
	 * Gets the path of the tools.jar file on the specified platform.
	 *
	 * @param platform a platform descriptor
	 * @param check true if the local presence of the tools.jar file should be checked
	 * @return the expected path of the tools.jar file on the specified platform, or null if it is not needed
	 * @throws FileNotFoundException if check is true and the path does not exist locally
	 */
	public static File getJavaToolsJarPath(PlatformDescriptor platform, boolean check) throws FileNotFoundException {

		if (platform == PlatformDescriptor.PLATFORM_MAC) return null;

		File path = new File(new File(getJavaSDKPath(platform), "lib"), TOOLS_JAR_NAME);

		if (check && !path.exists()) throw new FileNotFoundException("tools.jar path not found: " + path.getAbsolutePath());

		return path;
	}

	// ***********************************************************************

	/**
	 * Creates a shell command to invoke the Java interpreter with the specified class and arguments, using the class path of the currently executing program.
	 *
	 * @param clazz the class to be executed
	 * @param args the command line arguments
	 * @return a shell command to execute the class
	 * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
	 */
	public static String getJavaCommand(Class<?> clazz, List<String> args) throws FileNotFoundException {

		return getJavaCommand(clazz, args, ClassPath.getCurrentClassPath());
	}

	/**
	 * Creates a shell command to invoke the Java interpreter with the specified class and arguments, using the specified class path.
	 *
	 * @param clazz the class to be executed
	 * @param args the command line arguments
	 * @param class_path the class path
	 * @return a shell command to execute the class
	 * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
	 */
	public static String getJavaCommand(Class<?> clazz, List<String> args, ClassPath class_path) throws FileNotFoundException {

		return getJavaCommand(clazz, args, class_path, LOCAL_PLATFORM);
	}

	/**
	 * Creates a shell command to invoke the Java interpreter with the specified class and arguments on the specified platform, using the class path of the currently executing program.
	 *
	 * @param clazz the class to be executed
	 * @param args the command line arguments
	 * @param platform the platform
	 * @return a shell command to execute the class
	 * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
	 */
	public static String getJavaCommand(Class<?> clazz, List<String> args, PlatformDescriptor platform) throws FileNotFoundException {

		return getJavaCommand(clazz, args, new ClassPath(), platform);
	}

	/**
	 * Creates a shell command to invoke the Java interpreter with the specified class and arguments on the specified platform, using the specified class path.
	 *
	 * @param clazz the class to be executed
	 * @param args the command line arguments
	 * @param class_path the class path
	 * @param platform the platform
	 * @return a shell command to execute the class
	 * @throws FileNotFoundException if the Java interpreter executable cannot be found locally
	 */
	public static String getJavaCommand(Class<?> clazz, List<String> args, ClassPath class_path, PlatformDescriptor platform) throws FileNotFoundException {

		File interpreter_path = getJavaInterpreterPath(platform);
		String quote = getPathQuote(platform);

		StringBuffer buffer = new StringBuffer();

		buffer.append(quote);
		buffer.append(interpreter_path.getAbsolutePath());
		buffer.append(quote);

		buffer.append(" -cp ");
		buffer.append(class_path.toString(platform));

		buffer.append(" ");
		buffer.append(clazz.getName());

		for (String arg : args) {
			buffer.append(" ");
			buffer.append(arg);
		}

		return buffer.toString();
	}

	// ***********************************************************************

	/**
	 * Creates a shell command to invoke the Java compiler with the specified parameters.
	 *
	 * @param compilation_class_path the class path to be compiled against
	 * @param output_dir_path the path of the output directory
	 * @param compiler_flags the compiler flags
	 * @param file_paths the paths of the files to be compiled
	 * @return a shell command to invoke the compiler
	 * @throws FileNotFoundException if the compiler executable cannot be found locally
	 */
	public static String getJavaCompilerCommand(ClassPath compilation_class_path, File output_dir_path, List<String> compiler_flags, List<File> file_paths) throws FileNotFoundException {

		File compiler_command = getJavaCompilerPath();

		ClassPath current_class_path = ClassPath.getCurrentClassPath();
		current_class_path.append(compilation_class_path);

		StringBuffer buffer = new StringBuffer();

		buffer.append(getPathQuote());
		buffer.append(compiler_command.getAbsolutePath());
		buffer.append(getPathQuote());

		buffer.append(" -cp ");
		buffer.append(current_class_path.toString());

		buffer.append(" -d ");
		buffer.append(getPathQuote());
		buffer.append(output_dir_path.getAbsolutePath());
		buffer.append(getPathQuote());
		buffer.append(" ");

		for (String element : compiler_flags) {
			buffer.append(element);
			buffer.append(" ");
		}

		for (File element : file_paths) {
			buffer.append(getPathQuote());
			buffer.append(element.getAbsolutePath());
			buffer.append(getPathQuote());
			buffer.append(" ");
		}

		return buffer.toString();
	}

	/**
	 * Adds the local tools.jar path to the specified class path if it is needed on the current platform.
	 *
	 * @param class_path the class path
	 * @throws FileNotFoundException if the tools.jar file is needed on the current platform but it cannot be found
	 */
	public static void addJavaToolsJarIfNecessary(ClassPath class_path) throws FileNotFoundException {

		addJavaToolsJarIfNecessary(class_path, LOCAL_PLATFORM);
	}

	/**
	 * Adds the appropriate tools.jar path to the specified class path if it is needed on the specified platform.
	 *
	 * @param class_path the class path
	 * @param platform the platform
	 */
	public static void addJavaToolsJarIfNecessary(ClassPath class_path, PlatformDescriptor platform) {

		File tools_jar_path = getJavaToolsJarPath(platform);
		if (tools_jar_path != null) class_path.append(tools_jar_path);
	}

	/**
	 * @return the name of the main Java compiler class
	 */
	public static String getJavaCompilerMainClassName() {

		return StringConstants.getString("SystemUtil.JAVAC_MAIN_CLASS_NAME");
	}

	/**
	 * @return the name of the Java compiler output class
	 */
	public static String getJavaCompilerOutputClassName() {

		return StringConstants.getString("SystemUtil.OUTPUT_CLASS_NAME");
	}

	/**
	 * @return the name of the Java compiler executable on the local platform
	 */
	public static String getJavaCompilerName() {

		return getJavaCompilerName(LOCAL_PLATFORM);
	}

	/**
	 * @param platform a platform
	 * @return the name of the Java compiler executable on the specified platform
	 */
	public static String getJavaCompilerName(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_LINUX :   return "javac";
			case PLATFORM_MAC :     return "javac";
			case PLATFORM_WINDOWS : return "javac.exe";

			default : { ErrorHandling.hardError("unknown platform: " + platform); return null; }
		}
	}

	/**
	 * @param platform a platform
	 * @return the character, if any, to be used in quoting a file path on the specified platform
	 */
	public static String getPathQuote(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_LINUX :   return "";
			case PLATFORM_MAC :     return "";
			case PLATFORM_WINDOWS : return "\"";

			default : { ErrorHandling.hardError("unknown platform: " + platform); return null; }
		}
	}

	// ***********************************************************************

	private static String getJavaInterpreterName() {

		return getJavaInterpreterName(LOCAL_PLATFORM);
	}

	private static String getJavaInterpreterName(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_LINUX :   return "java";
			case PLATFORM_MAC :     return "java";
			case PLATFORM_WINDOWS : return "java.exe";

			default : { ErrorHandling.hardError("unknown platform: " + platform); return null; }
		}
	}

	private static String getJavaExecutableDirectoryName(PlatformDescriptor platform) {

		switch (platform) {
			case PLATFORM_LINUX :   return "bin";
			case PLATFORM_MAC :     return "Commands";
			case PLATFORM_WINDOWS : return "bin";

			default : { ErrorHandling.hardError("unknown platform: " + platform); return null; }
		}
	}

	private static String getPathQuote() {

		return getPathQuote(SystemUtil.LOCAL_PLATFORM);
	}
}
